#
# Copyright (C) 2007-2011 Edgewall Software, 2013-2025 the Babel team
# All rights reserved.
#
# This software is licensed as described in the file LICENSE, which
# you should have received as part of this distribution. The terms
# are also available at https://github.com/python-babel/babel/blob/master/LICENSE.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at https://github.com/python-babel/babel/commits/master/.

import calendar
from datetime import date, datetime, time, timedelta

import freezegun
import pytest

from babel import Locale, dates
from babel.dates import NO_INHERITANCE_MARKER, UTC, _localize, parse_pattern
from babel.util import FixedOffsetTimezone


class DateTimeFormatTestCase:

    def test_quarter_format(self):
        d = date(2006, 6, 8)
        fmt = dates.DateTimeFormat(d, locale='en_US')
        assert fmt['Q'] == '2'
        assert fmt['QQQQ'] == '2nd quarter'
        assert fmt['q'] == '2'
        assert fmt['qqqq'] == '2nd quarter'
        d = date(2006, 12, 31)
        fmt = dates.DateTimeFormat(d, locale='en_US')
        assert fmt['qqq'] == 'Q4'
        assert fmt['qqqqq'] == '4'
        assert fmt['QQQ'] == 'Q4'
        assert fmt['QQQQQ'] == '4'

    def test_month_context(self):
        d = date(2006, 2, 8)
        assert dates.DateTimeFormat(d, locale='mt_MT')['MMMMM'] == 'F'  # narrow format
        assert dates.DateTimeFormat(d, locale='mt_MT')['LLLLL'] == 'Fr'  # narrow standalone

    def test_abbreviated_month_alias(self):
        d = date(2006, 3, 8)
        assert dates.DateTimeFormat(d, locale='de_DE')['LLL'] == 'Mär'

    def test_week_of_year_first(self):
        d = date(2006, 1, 8)
        assert dates.DateTimeFormat(d, locale='de_DE')['w'] == '1'
        assert dates.DateTimeFormat(d, locale='en_US')['ww'] == '02'

    def test_week_of_year_first_with_year(self):
        d = date(2006, 1, 1)
        fmt = dates.DateTimeFormat(d, locale='de_DE')
        assert fmt['w'] == '52'
        assert fmt['YYYY'] == '2005'

    def test_week_of_year_last(self):
        d = date(2006, 12, 26)
        assert dates.DateTimeFormat(d, locale='de_DE')['w'] == '52'
        assert dates.DateTimeFormat(d, locale='en_US')['w'] == '52'

    def test_week_of_year_last_us_extra_week(self):
        d = date(2005, 12, 26)
        assert dates.DateTimeFormat(d, locale='de_DE')['w'] == '52'
        assert dates.DateTimeFormat(d, locale='en_US')['w'] == '53'

    def test_week_of_year_de_first_us_last_with_year(self):
        d = date(2018, 12, 31)
        fmt = dates.DateTimeFormat(d, locale='de_DE')
        assert fmt['w'] == '1'
        assert fmt['YYYY'] == '2019'
        fmt = dates.DateTimeFormat(d, locale='en_US')
        assert fmt['w'] == '53'
        assert fmt['yyyy'] == '2018'

    def test_week_of_month_first(self):
        d = date(2006, 1, 8)
        assert dates.DateTimeFormat(d, locale='de_DE')['W'] == '1'
        assert dates.DateTimeFormat(d, locale='en_US')['W'] == '2'

    def test_week_of_month_last(self):
        d = date(2006, 1, 29)
        assert dates.DateTimeFormat(d, locale='de_DE')['W'] == '4'
        assert dates.DateTimeFormat(d, locale='en_US')['W'] == '5'

    def test_day_of_year(self):
        d = date(2007, 4, 1)
        assert dates.DateTimeFormat(d, locale='en_US')['D'] == '91'

    def test_day_of_year_works_with_datetime(self):
        d = datetime(2007, 4, 1)
        assert dates.DateTimeFormat(d, locale='en_US')['D'] == '91'

    def test_day_of_year_first(self):
        d = date(2007, 1, 1)
        assert dates.DateTimeFormat(d, locale='en_US')['DDD'] == '001'

    def test_day_of_year_last(self):
        d = date(2007, 12, 31)
        assert dates.DateTimeFormat(d, locale='en_US')['DDD'] == '365'

    def test_day_of_week_in_month(self):
        d = date(2007, 4, 15)
        assert dates.DateTimeFormat(d, locale='en_US')['F'] == '3'

    def test_day_of_week_in_month_first(self):
        d = date(2007, 4, 1)
        assert dates.DateTimeFormat(d, locale='en_US')['F'] == '1'

    def test_day_of_week_in_month_last(self):
        d = date(2007, 4, 29)
        assert dates.DateTimeFormat(d, locale='en_US')['F'] == '5'

    def test_local_day_of_week(self):
        d = date(2007, 4, 1)  # a sunday
        assert dates.DateTimeFormat(d, locale='de_DE')['e'] == '7'  # monday is first day of week
        assert dates.DateTimeFormat(d, locale='en_US')['ee'] == '01'  # sunday is first day of week
        assert dates.DateTimeFormat(d, locale='ar_BH')['ee'] == '02'  # saturday is first day of week

        d = date(2007, 4, 2)  # a monday
        assert dates.DateTimeFormat(d, locale='de_DE')['e'] == '1'  # monday is first day of week
        assert dates.DateTimeFormat(d, locale='en_US')['ee'] == '02'  # sunday is first day of week
        assert dates.DateTimeFormat(d, locale='ar_BH')['ee'] == '03'  # saturday is first day of week

    def test_local_day_of_week_standalone(self):
        d = date(2007, 4, 1)  # a sunday
        assert dates.DateTimeFormat(d, locale='de_DE')['c'] == '7'  # monday is first day of week
        assert dates.DateTimeFormat(d, locale='en_US')['c'] == '1'  # sunday is first day of week
        assert dates.DateTimeFormat(d, locale='ar_BH')['c'] == '2'  # saturday is first day of week

        d = date(2007, 4, 2)  # a monday
        assert dates.DateTimeFormat(d, locale='de_DE')['c'] == '1'  # monday is first day of week
        assert dates.DateTimeFormat(d, locale='en_US')['c'] == '2'  # sunday is first day of week
        assert dates.DateTimeFormat(d, locale='ar_BH')['c'] == '3'  # saturday is first day of week

    def test_pattern_day_of_week(self):
        dt = datetime(2016, 2, 6)
        fmt = dates.DateTimeFormat(dt, locale='en_US')
        assert fmt['c'] == '7'
        assert fmt['ccc'] == 'Sat'
        assert fmt['cccc'] == 'Saturday'
        assert fmt['ccccc'] == 'S'
        assert fmt['cccccc'] == 'Sa'
        assert fmt['e'] == '7'
        assert fmt['ee'] == '07'
        assert fmt['eee'] == 'Sat'
        assert fmt['eeee'] == 'Saturday'
        assert fmt['eeeee'] == 'S'
        assert fmt['eeeeee'] == 'Sa'
        assert fmt['E'] == 'Sat'
        assert fmt['EE'] == 'Sat'
        assert fmt['EEE'] == 'Sat'
        assert fmt['EEEE'] == 'Saturday'
        assert fmt['EEEEE'] == 'S'
        assert fmt['EEEEEE'] == 'Sa'
        fmt = dates.DateTimeFormat(dt, locale='uk')
        assert fmt['c'] == '6'
        assert fmt['e'] == '6'
        assert fmt['ee'] == '06'

    def test_fractional_seconds(self):
        t = time(8, 3, 9, 799)
        assert dates.DateTimeFormat(t, locale='en_US')['S'] == '0'
        t = time(8, 3, 1, 799)
        assert dates.DateTimeFormat(t, locale='en_US')['SSSS'] == '0008'
        t = time(8, 3, 1, 34567)
        assert dates.DateTimeFormat(t, locale='en_US')['SSSS'] == '0346'
        t = time(8, 3, 1, 345678)
        assert dates.DateTimeFormat(t, locale='en_US')['SSSSSS'] == '345678'
        t = time(8, 3, 1, 799)
        assert dates.DateTimeFormat(t, locale='en_US')['SSSSS'] == '00080'

    def test_fractional_seconds_zero(self):
        t = time(15, 30, 0)
        assert dates.DateTimeFormat(t, locale='en_US')['SSSS'] == '0000'

    def test_milliseconds_in_day(self):
        t = time(15, 30, 12, 345000)
        assert dates.DateTimeFormat(t, locale='en_US')['AAAA'] == '55812345'

    def test_milliseconds_in_day_zero(self):
        d = time(0, 0, 0)
        assert dates.DateTimeFormat(d, locale='en_US')['AAAA'] == '0000'

    def test_timezone_rfc822(self, timezone_getter):
        tz = timezone_getter('Europe/Berlin')
        t = _localize(tz, datetime(2015, 1, 1, 15, 30))
        assert dates.DateTimeFormat(t, locale='de_DE')['Z'] == '+0100'

    def test_timezone_gmt(self, timezone_getter):
        tz = timezone_getter('Europe/Berlin')
        t = _localize(tz, datetime(2015, 1, 1, 15, 30))
        assert dates.DateTimeFormat(t, locale='de_DE')['ZZZZ'] == 'GMT+01:00'

    def test_timezone_name(self, timezone_getter):
        tz = timezone_getter('Europe/Paris')
        dt = _localize(tz, datetime(2007, 4, 1, 15, 30))
        assert dates.DateTimeFormat(dt, locale='fr_FR')['v'] == 'heure : France'

    def test_timezone_location_format(self, timezone_getter):
        tz = timezone_getter('Europe/Paris')
        dt = _localize(tz, datetime(2007, 4, 1, 15, 30))
        assert dates.DateTimeFormat(dt, locale='fr_FR')['VVVV'] == 'heure : France'

    def test_timezone_walltime_short(self, timezone_getter):
        tz = timezone_getter('Europe/Paris')
        t = time(15, 30, tzinfo=tz)
        assert dates.DateTimeFormat(t, locale='fr_FR')['v'] == 'heure : France'

    def test_timezone_walltime_long(self, timezone_getter):
        tz = timezone_getter('Europe/Paris')
        t = time(15, 30, tzinfo=tz)
        assert dates.DateTimeFormat(t, locale='fr_FR')['vvvv'] == 'heure d’Europe centrale'

    def test_hour_formatting(self):
        locale = 'en_US'
        t = time(0, 0, 0)
        assert dates.format_time(t, 'h a', locale=locale) == '12 AM'
        assert dates.format_time(t, 'H', locale=locale) == '0'
        assert dates.format_time(t, 'k', locale=locale) == '24'
        assert dates.format_time(t, 'K a', locale=locale) == '0 AM'
        t = time(12, 0, 0)
        assert dates.format_time(t, 'h a', locale=locale) == '12 PM'
        assert dates.format_time(t, 'H', locale=locale) == '12'
        assert dates.format_time(t, 'k', locale=locale) == '12'
        assert dates.format_time(t, 'K a', locale=locale) == '0 PM'


class FormatDateTestCase:

    def test_with_time_fields_in_pattern(self):
        with pytest.raises(AttributeError):
            dates.format_date(date(2007, 4, 1), "yyyy-MM-dd HH:mm", locale='en_US')

    def test_with_time_fields_in_pattern_and_datetime_param(self):
        with pytest.raises(AttributeError):
            dates.format_date(datetime(2007, 4, 1, 15, 30), "yyyy-MM-dd HH:mm", locale='en_US')

    def test_with_day_of_year_in_pattern_and_datetime_param(self):
        # format_date should work on datetimes just as well (see #282)
        d = datetime(2007, 4, 1)
        assert dates.format_date(d, 'w', locale='en_US') == '14'


class FormatDatetimeTestCase:

    def test_with_float(self, timezone_getter):
        UTC = timezone_getter('UTC')
        d = datetime(2012, 4, 1, 15, 30, 29, tzinfo=UTC)
        epoch = float(calendar.timegm(d.timetuple()))
        formatted_string = dates.format_datetime(epoch, format='long', locale='en_US')
        assert formatted_string == 'April 1, 2012 at 3:30:29 PM UTC'

    def test_timezone_formats_los_angeles(self, timezone_getter):
        tz = timezone_getter('America/Los_Angeles')
        dt = _localize(tz, datetime(2016, 1, 13, 7, 8, 35))
        assert dates.format_datetime(dt, 'z', locale='en') == 'PST'
        assert dates.format_datetime(dt, 'zz', locale='en') == 'PST'
        assert dates.format_datetime(dt, 'zzz', locale='en') == 'PST'
        assert dates.format_datetime(dt, 'zzzz', locale='en') == 'Pacific Standard Time'
        assert dates.format_datetime(dt, 'Z', locale='en') == '-0800'
        assert dates.format_datetime(dt, 'ZZ', locale='en') == '-0800'
        assert dates.format_datetime(dt, 'ZZZ', locale='en') == '-0800'
        assert dates.format_datetime(dt, 'ZZZZ', locale='en') == 'GMT-08:00'
        assert dates.format_datetime(dt, 'ZZZZZ', locale='en') == '-08:00'
        assert dates.format_datetime(dt, 'OOOO', locale='en') == 'GMT-08:00'
        assert dates.format_datetime(dt, 'VV', locale='en') == 'America/Los_Angeles'
        assert dates.format_datetime(dt, 'VVV', locale='en') == 'Los Angeles'
        assert dates.format_datetime(dt, 'X', locale='en') == '-08'
        assert dates.format_datetime(dt, 'XX', locale='en') == '-0800'
        assert dates.format_datetime(dt, 'XXX', locale='en') == '-08:00'
        assert dates.format_datetime(dt, 'XXXX', locale='en') == '-0800'
        assert dates.format_datetime(dt, 'XXXXX', locale='en') == '-08:00'
        assert dates.format_datetime(dt, 'x', locale='en') == '-08'
        assert dates.format_datetime(dt, 'xx', locale='en') == '-0800'
        assert dates.format_datetime(dt, 'xxx', locale='en') == '-08:00'
        assert dates.format_datetime(dt, 'xxxx', locale='en') == '-0800'
        assert dates.format_datetime(dt, 'xxxxx', locale='en') == '-08:00'

    def test_timezone_formats_utc(self, timezone_getter):
        tz = timezone_getter('UTC')
        dt = _localize(tz, datetime(2016, 1, 13, 7, 8, 35))
        assert dates.format_datetime(dt, 'Z', locale='en') == '+0000'
        assert dates.format_datetime(dt, 'ZZ', locale='en') == '+0000'
        assert dates.format_datetime(dt, 'ZZZ', locale='en') == '+0000'
        assert dates.format_datetime(dt, 'ZZZZ', locale='en') == 'GMT+00:00'
        assert dates.format_datetime(dt, 'ZZZZZ', locale='en') == 'Z'
        assert dates.format_datetime(dt, 'OOOO', locale='en') == 'GMT+00:00'
        assert dates.format_datetime(dt, 'VV', locale='en') == 'Etc/UTC'
        assert dates.format_datetime(dt, 'VVV', locale='en') == 'UTC'
        assert dates.format_datetime(dt, 'X', locale='en') == 'Z'
        assert dates.format_datetime(dt, 'XX', locale='en') == 'Z'
        assert dates.format_datetime(dt, 'XXX', locale='en') == 'Z'
        assert dates.format_datetime(dt, 'XXXX', locale='en') == 'Z'
        assert dates.format_datetime(dt, 'XXXXX', locale='en') == 'Z'
        assert dates.format_datetime(dt, 'x', locale='en') == '+00'
        assert dates.format_datetime(dt, 'xx', locale='en') == '+0000'
        assert dates.format_datetime(dt, 'xxx', locale='en') == '+00:00'
        assert dates.format_datetime(dt, 'xxxx', locale='en') == '+0000'
        assert dates.format_datetime(dt, 'xxxxx', locale='en') == '+00:00'

    def test_timezone_formats_kolkata(self, timezone_getter):
        tz = timezone_getter('Asia/Kolkata')
        dt = _localize(tz, datetime(2016, 1, 13, 7, 8, 35))
        assert dates.format_datetime(dt, 'zzzz', locale='en') == 'India Standard Time'
        assert dates.format_datetime(dt, 'ZZZZ', locale='en') == 'GMT+05:30'
        assert dates.format_datetime(dt, 'ZZZZZ', locale='en') == '+05:30'
        assert dates.format_datetime(dt, 'OOOO', locale='en') == 'GMT+05:30'
        assert dates.format_datetime(dt, 'VV', locale='en') == 'Asia/Calcutta'
        assert dates.format_datetime(dt, 'VVV', locale='en') == 'Kolkata'
        assert dates.format_datetime(dt, 'X', locale='en') == '+0530'
        assert dates.format_datetime(dt, 'XX', locale='en') == '+0530'
        assert dates.format_datetime(dt, 'XXX', locale='en') == '+05:30'
        assert dates.format_datetime(dt, 'XXXX', locale='en') == '+0530'
        assert dates.format_datetime(dt, 'XXXXX', locale='en') == '+05:30'
        assert dates.format_datetime(dt, 'x', locale='en') == '+0530'
        assert dates.format_datetime(dt, 'xx', locale='en') == '+0530'
        assert dates.format_datetime(dt, 'xxx', locale='en') == '+05:30'
        assert dates.format_datetime(dt, 'xxxx', locale='en') == '+0530'
        assert dates.format_datetime(dt, 'xxxxx', locale='en') == '+05:30'


class FormatTimeTestCase:

    def test_with_naive_datetime_and_tzinfo(self, timezone_getter):
        assert dates.format_time(
            datetime(2007, 4, 1, 15, 30),
            'long',
            tzinfo=timezone_getter('US/Eastern'),
            locale='en',
        ) == '11:30:00 AM EDT'

    def test_with_float(self, timezone_getter):
        tz = timezone_getter('UTC')
        d = _localize(tz, datetime(2012, 4, 1, 15, 30, 29))
        epoch = float(calendar.timegm(d.timetuple()))
        assert dates.format_time(epoch, format='long', locale='en_US') == '3:30:29 PM UTC'

    def test_with_date_fields_in_pattern(self):
        with pytest.raises(AttributeError):
            dates.format_time(datetime(2007, 4, 1), 'yyyy-MM-dd HH:mm', locale='en')

    def test_with_date_fields_in_pattern_and_datetime_param(self):
        with pytest.raises(AttributeError):
            dates.format_time(datetime(2007, 4, 1, 15, 30), "yyyy-MM-dd HH:mm", locale='en_US')


class FormatTimedeltaTestCase:

    def test_zero_seconds(self):
        td = timedelta(seconds=0)
        assert dates.format_timedelta(td, locale='en') == '0 seconds'
        assert dates.format_timedelta(td, locale='en', format='short') == '0 sec'
        assert dates.format_timedelta(td, granularity='hour', locale='en') == '0 hours'
        assert dates.format_timedelta(td, granularity='hour', locale='en', format='short') == '0 hr'

    def test_small_value_with_granularity(self):
        td = timedelta(seconds=42)
        assert dates.format_timedelta(td, granularity='hour', locale='en') == '1 hour'
        assert dates.format_timedelta(td, granularity='hour', locale='en', format='short') == '1 hr'

    def test_direction_adding(self):
        td = timedelta(hours=1)
        assert dates.format_timedelta(td, locale='en', add_direction=True) == 'in 1 hour'
        assert dates.format_timedelta(-td, locale='en', add_direction=True) == '1 hour ago'

    def test_format_narrow(self):
        assert dates.format_timedelta(timedelta(hours=1), locale='en', format='narrow') == '1h'
        assert dates.format_timedelta(timedelta(hours=-2), locale='en', format='narrow') == '2h'

    def test_format_invalid(self):
        for format in (None, '', 'bold italic'):
            with pytest.raises(TypeError):
                dates.format_timedelta(timedelta(hours=1), format=format)


class TimeZoneAdjustTestCase:

    def _utc(self):
        class EvilFixedOffsetTimezone(FixedOffsetTimezone):

            def localize(self, dt, is_dst=False):
                raise NotImplementedError()
        UTC = EvilFixedOffsetTimezone(0, 'UTC')
        # This is important to trigger the actual bug (#257)
        assert hasattr(UTC, 'normalize') is False
        return UTC

    def test_can_format_time_with_custom_timezone(self):
        # regression test for #257
        utc = self._utc()
        t = datetime(2007, 4, 1, 15, 30, tzinfo=utc)
        formatted_time = dates.format_time(t, 'long', tzinfo=utc, locale='en')
        assert formatted_time == '3:30:00 PM UTC'


def test_get_period_names():
    assert dates.get_period_names(locale='en_US')['am'] == 'AM'


def test_get_day_names():
    assert dates.get_day_names('wide', locale='en_US')[1] == 'Tuesday'
    assert dates.get_day_names('short', locale='en_US')[1] == 'Tu'
    assert dates.get_day_names('abbreviated', locale='es')[1] == 'mar'
    de = dates.get_day_names('narrow', context='stand-alone', locale='de_DE')
    assert de[1] == 'D'


def test_get_month_names():
    assert dates.get_month_names('wide', locale='en_US')[1] == 'January'
    assert dates.get_month_names('abbreviated', locale='es')[1] == 'ene'
    de = dates.get_month_names('narrow', context='stand-alone', locale='de_DE')
    assert de[1] == 'J'


def test_get_quarter_names():
    assert dates.get_quarter_names('wide', locale='en_US')[1] == '1st quarter'
    assert dates.get_quarter_names('abbreviated', locale='de_DE')[1] == 'Q1'
    assert dates.get_quarter_names('narrow', locale='de_DE')[1] == '1'


def test_get_era_names():
    assert dates.get_era_names('wide', locale='en_US')[1] == 'Anno Domini'
    assert dates.get_era_names('abbreviated', locale='de_DE')[1] == 'n. Chr.'


def test_get_date_format():
    us = dates.get_date_format(locale='en_US')
    assert us.pattern == 'MMM d, y'
    de = dates.get_date_format('full', locale='de_DE')
    assert de.pattern == 'EEEE, d. MMMM y'


def test_get_datetime_format():
    assert dates.get_datetime_format(locale='en_US') == '{1}, {0}'


def test_get_time_format():
    assert dates.get_time_format(locale='en_US').pattern == 'h:mm:ss\u202fa'
    assert (dates.get_time_format('full', locale='de_DE').pattern ==
            'HH:mm:ss zzzz')


def test_get_timezone_gmt(timezone_getter):
    dt = datetime(2007, 4, 1, 15, 30)
    assert dates.get_timezone_gmt(dt, locale='en') == 'GMT+00:00'
    assert dates.get_timezone_gmt(dt, locale='en', return_z=True) == 'Z'
    assert dates.get_timezone_gmt(dt, locale='en', width='iso8601_short') == '+00'
    tz = timezone_getter('America/Los_Angeles')
    dt = _localize(tz, datetime(2007, 4, 1, 15, 30))
    assert dates.get_timezone_gmt(dt, locale='en') == 'GMT-07:00'
    assert dates.get_timezone_gmt(dt, 'short', locale='en') == '-0700'
    assert dates.get_timezone_gmt(dt, locale='en', width='iso8601_short') == '-07'
    assert dates.get_timezone_gmt(dt, 'long', locale='fr_FR') == 'UTC-07:00'


def test_get_timezone_location(timezone_getter):
    tz = timezone_getter('America/St_Johns')
    assert (dates.get_timezone_location(tz, locale='de_DE') ==
            "Kanada (St. John\u2019s) (Ortszeit)")
    assert (dates.get_timezone_location(tz, locale='en') ==
            'Canada (St. John’s) Time')
    assert (dates.get_timezone_location(tz, locale='en', return_city=True) ==
            'St. John’s')

    tz = timezone_getter('America/Mexico_City')
    assert (dates.get_timezone_location(tz, locale='de_DE') ==
            'Mexiko (Mexiko-Stadt) (Ortszeit)')

    tz = timezone_getter('Europe/Berlin')
    assert (dates.get_timezone_location(tz, locale='de_DE') ==
            'Deutschland (Berlin) (Ortszeit)')


@pytest.mark.parametrize(
    "tzname, params, expected",
    [
        ("America/Los_Angeles", {"locale": "en_US"}, "Pacific Time"),
        ("America/Los_Angeles", {"width": "short", "locale": "en_US"}, "PT"),
        ("Europe/Berlin", {"locale": "de_DE"}, "Mitteleurop\xe4ische Zeit"),
        ("Europe/Berlin", {"locale": "pt_BR"}, "Hor\xe1rio da Europa Central"),
        ("America/St_Johns", {"locale": "de_DE"}, "Neufundland-Zeit"),
        (
            "America/Los_Angeles",
            {"locale": "en", "width": "short", "zone_variant": "generic"},
            "PT",
        ),
        (
            "America/Los_Angeles",
            {"locale": "en", "width": "short", "zone_variant": "standard"},
            "PST",
        ),
        (
            "America/Los_Angeles",
            {"locale": "en", "width": "short", "zone_variant": "daylight"},
            "PDT",
        ),
        (
            "America/Los_Angeles",
            {"locale": "en", "width": "long", "zone_variant": "generic"},
            "Pacific Time",
        ),
        (
            "America/Los_Angeles",
            {"locale": "en", "width": "long", "zone_variant": "standard"},
            "Pacific Standard Time",
        ),
        (
            "America/Los_Angeles",
            {"locale": "en", "width": "long", "zone_variant": "daylight"},
            "Pacific Daylight Time",
        ),
        ("Europe/Berlin", {"locale": "en_US"}, "Central European Time"),
    ],
)
def test_get_timezone_name_tzinfo(timezone_getter, tzname, params, expected):
    tz = timezone_getter(tzname)
    assert dates.get_timezone_name(tz, **params) == expected


@pytest.mark.parametrize("timezone_getter", ["pytz.timezone"], indirect=True)
@pytest.mark.parametrize(
    "tzname, params, expected",
    [
        ("America/Los_Angeles", {"locale": "en_US"}, "Pacific Standard Time"),
        (
            "America/Los_Angeles",
            {"locale": "en_US", "return_zone": True},
            "America/Los_Angeles",
        ),
        ("America/Los_Angeles", {"width": "short", "locale": "en_US"}, "PST"),
    ],
)
def test_get_timezone_name_time_pytz(timezone_getter, tzname, params, expected):
    """pytz (by design) can't determine if the time is in DST or not,
    so it will always return Standard time"""
    dt = time(15, 30, tzinfo=timezone_getter(tzname))
    assert dates.get_timezone_name(dt, **params) == expected


def test_get_timezone_name_misc(timezone_getter):
    localnow = datetime.now(timezone_getter('UTC')).astimezone(dates.LOCALTZ)
    assert (dates.get_timezone_name(None, locale='en_US') ==
            dates.get_timezone_name(localnow, locale='en_US'))

    assert (dates.get_timezone_name('Europe/Berlin', locale='en_US') == "Central European Time")

    assert (dates.get_timezone_name(1400000000, locale='en_US', width='short') == "Unknown Region (UTC) Time")
    assert (dates.get_timezone_name(time(16, 20), locale='en_US', width='short') == "UTC")


def test_format_date():
    d = date(2007, 4, 1)
    assert dates.format_date(d, locale='en_US') == 'Apr 1, 2007'
    assert (dates.format_date(d, format='full', locale='de_DE') ==
            'Sonntag, 1. April 2007')
    assert (dates.format_date(d, "EEE, MMM d, ''yy", locale='en') ==
            "Sun, Apr 1, '07")


def test_format_datetime(timezone_getter):
    dt = datetime(2007, 4, 1, 15, 30)
    assert (dates.format_datetime(dt, locale='en_US') ==
            'Apr 1, 2007, 3:30:00\u202fPM')

    full = dates.format_datetime(
        dt, 'full',
        tzinfo=timezone_getter('Europe/Paris'),
        locale='fr_FR',
    )
    assert full == (
        'dimanche 1 avril 2007, 17:30:00 heure '
        'd’été d’Europe centrale'
    )
    custom = dates.format_datetime(
        dt, "yyyy.MM.dd G 'at' HH:mm:ss zzz",
        tzinfo=timezone_getter('US/Eastern'),
        locale='en',
    )
    assert custom == '2007.04.01 AD at 11:30:00 EDT'


def test_format_time(timezone_getter):
    t = time(15, 30)
    assert dates.format_time(t, locale='en_US') == '3:30:00\u202fPM'
    assert dates.format_time(t, format='short', locale='de_DE') == '15:30'

    assert (dates.format_time(t, "hh 'o''clock' a", locale='en') ==
            "03 o'clock PM")

    paris = timezone_getter('Europe/Paris')
    eastern = timezone_getter('US/Eastern')

    t = _localize(paris, datetime(2007, 4, 1, 15, 30))
    fr = dates.format_time(t, format='full', tzinfo=paris, locale='fr_FR')
    assert fr == '15:30:00 heure d’été d’Europe centrale'

    custom = dates.format_time(t, "hh 'o''clock' a, zzzz", tzinfo=eastern, locale='en')
    assert custom == "09 o'clock AM, Eastern Daylight Time"

    with freezegun.freeze_time("2023-01-01"):
        t = time(15, 30)
        paris = dates.format_time(t, format='full', tzinfo=paris, locale='fr_FR')
        assert paris == '15:30:00 heure normale d’Europe centrale'

        us_east = dates.format_time(t, format='full', tzinfo=eastern, locale='en_US')
        assert us_east == '3:30:00\u202fPM Eastern Standard Time'


def test_format_skeleton(timezone_getter):
    dt = datetime(2007, 4, 1, 15, 30)
    assert (dates.format_skeleton('yMEd', dt, locale='en_US') == 'Sun, 4/1/2007')
    assert (dates.format_skeleton('yMEd', dt, locale='th') == 'อา. 1/4/2007')

    assert (dates.format_skeleton('EHm', dt, locale='en') == 'Sun 15:30')
    assert (dates.format_skeleton('EHm', dt, tzinfo=timezone_getter('Asia/Bangkok'), locale='th') == 'อา. 22:30 น.')


@pytest.mark.parametrize(('skeleton', 'expected'), [
    ('Hmz', 'Hmv'),
    ('kmv', 'Hmv'),
    ('Kmv', 'hmv'),
    ('Hma', 'Hm'),
    ('Hmb', 'Hm'),
    ('zkKab', 'vHh'),
])
def test_match_skeleton_alternate_characters(skeleton, expected):
    # https://github.com/unicode-org/icu/blob/5e22f0076ec9b55056cd8a84e9ef370632f44174/icu4j/main/core/src/main/java/com/ibm/icu/text/DateIntervalInfo.java#L1090-L1102
    assert dates.match_skeleton(skeleton, (expected,)) == expected


def test_format_timedelta():
    assert (dates.format_timedelta(timedelta(weeks=12), locale='en_US')
            == '3 months')
    assert (dates.format_timedelta(timedelta(seconds=1), locale='es')
            == '1 segundo')

    assert (dates.format_timedelta(timedelta(hours=3), granularity='day',
                                   locale='en_US')
            == '1 day')

    assert (dates.format_timedelta(timedelta(hours=23), threshold=0.9,
                                   locale='en_US')
            == '1 day')
    assert (dates.format_timedelta(timedelta(hours=23), threshold=1.1,
                                   locale='en_US')
            == '23 hours')


def test_parse_date():
    assert dates.parse_date('4/1/04', locale='en_US') == date(2004, 4, 1)
    assert dates.parse_date('01.04.2004', locale='de_DE') == date(2004, 4, 1)
    assert dates.parse_date('2004-04-01', locale='sv_SE', format='short') == date(2004, 4, 1)


def test_parse_date_custom_format():
    assert dates.parse_date('1.4.2024', format='dd.mm.yyyy') == date(2024, 4, 1)
    assert dates.parse_date('2024.4.1', format='yyyy.mm.dd') == date(2024, 4, 1)
    # Dates that look like ISO 8601 should use the custom format as well:
    assert dates.parse_date('2024-04-01', format='yyyy.dd.mm') == date(2024, 1, 4)


@pytest.mark.parametrize('input, expected', [
    # base case, fully qualified time
    ('15:30:00', time(15, 30)),
    # test digits
    ('15:30', time(15, 30)),
    ('3:30', time(3, 30)),
    ('00:30', time(0, 30)),
    # test am parsing
    ('03:30 am', time(3, 30)),
    ('3:30:21 am', time(3, 30, 21)),
    ('3:30 am', time(3, 30)),
    # test pm parsing
    ('03:30 pm', time(15, 30)),
    ('03:30 pM', time(15, 30)),
    ('03:30 Pm', time(15, 30)),
    ('03:30 PM', time(15, 30)),
    # test hour-only parsing
    ('4 pm', time(16, 0)),
])
def test_parse_time(input, expected):
    assert dates.parse_time(input, locale='en_US') == expected


def test_parse_time_no_seconds_in_format():
    # parse time using a time format which does not include seconds
    locale = 'cs_CZ'
    fmt = 'short'
    assert dates.get_time_format(format=fmt, locale=locale).pattern == 'H:mm'
    assert dates.parse_time('9:30', locale=locale, format=fmt) == time(9, 30)


def test_parse_time_alternate_characters(monkeypatch):
    # 'K' can be used as an alternative to 'H'
    def get_time_format(*args, **kwargs):
        return parse_pattern('KK:mm:ss')
    monkeypatch.setattr(dates, 'get_time_format', get_time_format)

    assert dates.parse_time('9:30') == time(9, 30)


def test_parse_date_alternate_characters(monkeypatch):
    # 'l' can be used as an alternative to 'm'
    def get_date_format(*args, **kwargs):
        return parse_pattern('yyyy-ll-dd')
    monkeypatch.setattr(dates, 'get_time_format', get_date_format)

    assert dates.parse_date('2024-10-20') == date(2024, 10, 20)


def test_parse_time_custom_format():
    assert dates.parse_time('15:30:00', format='HH:mm:ss') == time(15, 30)
    assert dates.parse_time('00:30:15', format='ss:mm:HH') == time(15, 30)


@pytest.mark.parametrize('case', ['', 'a', 'aaa'])
@pytest.mark.parametrize('func', [dates.parse_date, dates.parse_time])
def test_parse_errors(case, func):
    with pytest.raises(dates.ParseError):
        func(case, locale='en_US')


def test_datetime_format_get_week_number():
    format = dates.DateTimeFormat(date(2006, 1, 8), Locale.parse('de_DE'))
    assert format.get_week_number(6) == 1

    format = dates.DateTimeFormat(date(2006, 1, 8), Locale.parse('en_US'))
    assert format.get_week_number(6) == 2


def test_parse_pattern():
    assert dates.parse_pattern("MMMMd").format == '%(MMMM)s%(d)s'
    assert (dates.parse_pattern("MMM d, yyyy").format ==
            '%(MMM)s %(d)s, %(yyyy)s')
    assert (dates.parse_pattern("H:mm' Uhr 'z").format ==
            '%(H)s:%(mm)s Uhr %(z)s')
    assert dates.parse_pattern("hh' o''clock'").format == "%(hh)s o'clock"


def test_lithuanian_long_format():
    assert (
        dates.format_date(date(2015, 12, 10), locale='lt_LT', format='long') ==
        '2015 m. gruodžio 10 d.'
    )


def test_zh_TW_format():
    # Refs GitHub issue #378
    assert dates.format_time(datetime(2016, 4, 8, 12, 34, 56), locale='zh_TW') == '中午12:34:56'


def test_format_current_moment():
    frozen_instant = datetime.now(UTC)
    with freezegun.freeze_time(time_to_freeze=frozen_instant):
        assert dates.format_datetime(locale="en_US") == dates.format_datetime(frozen_instant, locale="en_US")


@pytest.mark.all_locales
def test_no_inherit_metazone_marker_never_in_output(locale, timezone_getter):
    # See: https://github.com/python-babel/babel/issues/428
    tz = timezone_getter('America/Los_Angeles')
    t = _localize(tz, datetime(2016, 1, 6, 7))
    assert NO_INHERITANCE_MARKER not in dates.format_time(t, format='long', locale=locale)
    assert NO_INHERITANCE_MARKER not in dates.get_timezone_name(t, width='short', locale=locale)


def test_no_inherit_metazone_formatting(timezone_getter):
    # See: https://github.com/python-babel/babel/issues/428
    tz = timezone_getter('America/Los_Angeles')
    t = _localize(tz, datetime(2016, 1, 6, 7))
    assert dates.format_time(t, format='long', locale='en_US') == "7:00:00\u202fAM PST"
    assert dates.format_time(t, format='long', locale='en_GB') == "07:00:00 Pacific Standard Time"
    assert dates.get_timezone_name(t, width='short', locale='en_US') == "PST"
    assert dates.get_timezone_name(t, width='short', locale='en_GB') == "Pacific Standard Time"


def test_russian_week_numbering():
    # See https://github.com/python-babel/babel/issues/485
    v = date(2017, 1, 1)
    assert dates.format_date(v, format='YYYY-ww', locale='ru_RU') == '2016-52'  # This would have returned 2017-01 prior to CLDR 32
    assert dates.format_date(v, format='YYYY-ww', locale='de_DE') == '2016-52'


def test_week_numbering_isocalendar():
    locale = Locale.parse('de_DE')
    assert locale.first_week_day == 0
    assert locale.min_week_days == 4

    def week_number(value):
        return dates.format_date(value, format="YYYY-'W'ww-e", locale=locale)

    for year in range(1991, 2010):
        for delta in range(-9, 9):
            value = date(year, 1, 1) + timedelta(days=delta)
            expected = '%04d-W%02d-%d' % value.isocalendar()
            assert week_number(value) == expected

def test_week_numbering_monday_mindays_4():
    locale = Locale.parse('de_DE')
    assert locale.first_week_day == 0
    assert locale.min_week_days == 4

    def week_number(value):
        value = date.fromisoformat(value)
        return dates.format_date(value, format="YYYY-'W'ww-e", locale=locale)

    def week_of_month(value):
        value = date.fromisoformat(value)
        return dates.format_date(value, format='W', locale=locale)

    assert week_number('2003-01-01') == '2003-W01-3'
    assert week_number('2003-01-06') == '2003-W02-1'
    assert week_number('2003-12-28') == '2003-W52-7'
    assert week_number('2003-12-29') == '2004-W01-1'
    assert week_number('2003-12-31') == '2004-W01-3'
    assert week_number('2004-01-01') == '2004-W01-4'
    assert week_number('2004-01-05') == '2004-W02-1'
    assert week_number('2004-12-26') == '2004-W52-7'
    assert week_number('2004-12-27') == '2004-W53-1'
    assert week_number('2004-12-31') == '2004-W53-5'
    assert week_number('2005-01-01') == '2004-W53-6'
    assert week_number('2005-01-03') == '2005-W01-1'
    assert week_number('2005-01-10') == '2005-W02-1'
    assert week_number('2005-12-31') == '2005-W52-6'
    assert week_number('2006-01-01') == '2005-W52-7'
    assert week_number('2006-01-02') == '2006-W01-1'
    assert week_number('2006-01-09') == '2006-W02-1'
    assert week_number('2006-12-31') == '2006-W52-7'
    assert week_number('2007-01-01') == '2007-W01-1'
    assert week_number('2007-12-30') == '2007-W52-7'
    assert week_number('2007-12-31') == '2008-W01-1'
    assert week_number('2008-01-01') == '2008-W01-2'
    assert week_number('2008-01-07') == '2008-W02-1'
    assert week_number('2008-12-28') == '2008-W52-7'
    assert week_number('2008-12-29') == '2009-W01-1'
    assert week_number('2008-12-31') == '2009-W01-3'
    assert week_number('2009-01-01') == '2009-W01-4'
    assert week_number('2009-01-05') == '2009-W02-1'
    assert week_number('2009-12-27') == '2009-W52-7'
    assert week_number('2009-12-28') == '2009-W53-1'
    assert week_number('2009-12-31') == '2009-W53-4'
    assert week_number('2010-01-01') == '2009-W53-5'
    assert week_number('2010-01-03') == '2009-W53-7'
    assert week_number('2010-01-04') == '2010-W01-1'

    assert week_of_month('2003-01-01') == '1'  # Wed
    assert week_of_month('2003-02-28') == '4'
    assert week_of_month('2003-05-01') == '1'  # Thu
    assert week_of_month('2003-08-01') == '0'  # Fri
    assert week_of_month('2003-12-31') == '5'
    assert week_of_month('2004-01-01') == '1'  # Thu
    assert week_of_month('2004-02-29') == '4'


def test_week_numbering_monday_mindays_1():
    locale = Locale.parse('tr_TR')
    assert locale.first_week_day == 0
    assert locale.min_week_days == 1

    def week_number(value):
        value = date.fromisoformat(value)
        return dates.format_date(value, format="YYYY-'W'ww-e", locale=locale)

    def week_of_month(value):
        value = date.fromisoformat(value)
        return dates.format_date(value, format='W', locale=locale)

    assert week_number('2003-01-01') == '2003-W01-3'
    assert week_number('2003-01-06') == '2003-W02-1'
    assert week_number('2003-12-28') == '2003-W52-7'
    assert week_number('2003-12-29') == '2004-W01-1'
    assert week_number('2003-12-31') == '2004-W01-3'
    assert week_number('2004-01-01') == '2004-W01-4'
    assert week_number('2004-01-05') == '2004-W02-1'
    assert week_number('2004-12-26') == '2004-W52-7'
    assert week_number('2004-12-27') == '2005-W01-1'
    assert week_number('2004-12-31') == '2005-W01-5'
    assert week_number('2005-01-01') == '2005-W01-6'
    assert week_number('2005-01-03') == '2005-W02-1'
    assert week_number('2005-12-25') == '2005-W52-7'
    assert week_number('2005-12-26') == '2006-W01-1'
    assert week_number('2005-12-31') == '2006-W01-6'
    assert week_number('2006-01-01') == '2006-W01-7'
    assert week_number('2006-01-02') == '2006-W02-1'
    assert week_number('2006-12-31') == '2006-W53-7'
    assert week_number('2007-01-01') == '2007-W01-1'
    assert week_number('2007-01-08') == '2007-W02-1'
    assert week_number('2007-12-30') == '2007-W52-7'
    assert week_number('2007-12-31') == '2008-W01-1'
    assert week_number('2008-01-01') == '2008-W01-2'
    assert week_number('2008-01-07') == '2008-W02-1'
    assert week_number('2008-12-28') == '2008-W52-7'
    assert week_number('2008-12-29') == '2009-W01-1'
    assert week_number('2008-12-31') == '2009-W01-3'
    assert week_number('2009-01-01') == '2009-W01-4'
    assert week_number('2009-01-05') == '2009-W02-1'
    assert week_number('2009-12-27') == '2009-W52-7'
    assert week_number('2009-12-28') == '2010-W01-1'
    assert week_number('2009-12-31') == '2010-W01-4'
    assert week_number('2010-01-01') == '2010-W01-5'
    assert week_number('2010-01-03') == '2010-W01-7'
    assert week_number('2010-01-04') == '2010-W02-1'

    assert week_of_month('2003-01-01') == '1'  # Wed
    assert week_of_month('2003-02-28') == '5'
    assert week_of_month('2003-06-01') == '1'  # Sun
    assert week_of_month('2003-12-31') == '5'
    assert week_of_month('2004-01-01') == '1'  # Thu
    assert week_of_month('2004-02-29') == '5'


def test_week_numbering_sunday_mindays_1():
    locale = Locale.parse('en_US')
    assert locale.first_week_day == 6
    assert locale.min_week_days == 1

    def week_number(value):
        value = date.fromisoformat(value)
        return dates.format_date(value, format="YYYY-'W'ww-e", locale=locale)

    def week_of_month(value):
        value = date.fromisoformat(value)
        return dates.format_date(value, format='W', locale=locale)

    assert week_number('2003-01-01') == '2003-W01-4'
    assert week_number('2003-01-05') == '2003-W02-1'
    assert week_number('2003-12-27') == '2003-W52-7'
    assert week_number('2003-12-28') == '2004-W01-1'
    assert week_number('2003-12-31') == '2004-W01-4'
    assert week_number('2004-01-01') == '2004-W01-5'
    assert week_number('2004-01-04') == '2004-W02-1'
    assert week_number('2004-12-25') == '2004-W52-7'
    assert week_number('2004-12-26') == '2005-W01-1'
    assert week_number('2004-12-31') == '2005-W01-6'
    assert week_number('2005-01-01') == '2005-W01-7'
    assert week_number('2005-01-02') == '2005-W02-1'
    assert week_number('2005-12-24') == '2005-W52-7'
    assert week_number('2005-12-25') == '2005-W53-1'
    assert week_number('2005-12-31') == '2005-W53-7'
    assert week_number('2006-01-01') == '2006-W01-1'
    assert week_number('2006-01-08') == '2006-W02-1'
    assert week_number('2006-12-30') == '2006-W52-7'
    assert week_number('2006-12-31') == '2007-W01-1'
    assert week_number('2007-01-01') == '2007-W01-2'
    assert week_number('2007-01-07') == '2007-W02-1'
    assert week_number('2007-12-29') == '2007-W52-7'
    assert week_number('2007-12-30') == '2008-W01-1'
    assert week_number('2007-12-31') == '2008-W01-2'
    assert week_number('2008-01-01') == '2008-W01-3'
    assert week_number('2008-01-06') == '2008-W02-1'
    assert week_number('2008-12-27') == '2008-W52-7'
    assert week_number('2008-12-28') == '2009-W01-1'
    assert week_number('2008-12-31') == '2009-W01-4'
    assert week_number('2009-01-01') == '2009-W01-5'
    assert week_number('2009-01-04') == '2009-W02-1'
    assert week_number('2009-12-26') == '2009-W52-7'
    assert week_number('2009-12-27') == '2010-W01-1'
    assert week_number('2009-12-31') == '2010-W01-5'
    assert week_number('2010-01-01') == '2010-W01-6'
    assert week_number('2010-01-03') == '2010-W02-1'

    assert week_of_month('2003-01-01') == '1'  # Wed
    assert week_of_month('2003-02-01') == '1'  # Sat
    assert week_of_month('2003-02-28') == '5'
    assert week_of_month('2003-12-31') == '5'
    assert week_of_month('2004-01-01') == '1'  # Thu
    assert week_of_month('2004-02-29') == '5'


def test_week_numbering_sunday_mindays_4():
    locale = Locale.parse('pt_PT')
    assert locale.first_week_day == 6
    assert locale.min_week_days == 4

    def week_number(value):
        value = date.fromisoformat(value)
        return dates.format_date(value, format="YYYY-'W'ww-e", locale=locale)

    def week_of_month(value):
        value = date.fromisoformat(value)
        return dates.format_date(value, format='W', locale=locale)

    assert week_number('2003-01-01') == '2003-W01-4'
    assert week_number('2003-01-05') == '2003-W02-1'
    assert week_number('2003-12-27') == '2003-W52-7'
    assert week_number('2003-12-28') == '2003-W53-1'
    assert week_number('2003-12-31') == '2003-W53-4'
    assert week_number('2004-01-01') == '2003-W53-5'
    assert week_number('2004-01-04') == '2004-W01-1'
    assert week_number('2004-12-25') == '2004-W51-7'
    assert week_number('2004-12-26') == '2004-W52-1'
    assert week_number('2004-12-31') == '2004-W52-6'
    assert week_number('2005-01-01') == '2004-W52-7'
    assert week_number('2005-01-02') == '2005-W01-1'
    assert week_number('2005-12-24') == '2005-W51-7'
    assert week_number('2005-12-25') == '2005-W52-1'
    assert week_number('2005-12-31') == '2005-W52-7'
    assert week_number('2006-01-01') == '2006-W01-1'
    assert week_number('2006-01-08') == '2006-W02-1'
    assert week_number('2006-12-30') == '2006-W52-7'
    assert week_number('2006-12-31') == '2007-W01-1'
    assert week_number('2007-01-01') == '2007-W01-2'
    assert week_number('2007-01-07') == '2007-W02-1'
    assert week_number('2007-12-29') == '2007-W52-7'
    assert week_number('2007-12-30') == '2008-W01-1'
    assert week_number('2007-12-31') == '2008-W01-2'
    assert week_number('2008-01-01') == '2008-W01-3'
    assert week_number('2008-01-06') == '2008-W02-1'
    assert week_number('2008-12-27') == '2008-W52-7'
    assert week_number('2008-12-28') == '2008-W53-1'
    assert week_number('2008-12-31') == '2008-W53-4'
    assert week_number('2009-01-01') == '2008-W53-5'
    assert week_number('2009-01-04') == '2009-W01-1'
    assert week_number('2009-12-26') == '2009-W51-7'
    assert week_number('2009-12-27') == '2009-W52-1'
    assert week_number('2009-12-31') == '2009-W52-5'
    assert week_number('2010-01-01') == '2009-W52-6'
    assert week_number('2010-01-03') == '2010-W01-1'

    assert week_of_month('2003-01-01') == '1'  # Wed
    assert week_of_month('2003-02-28') == '4'
    assert week_of_month('2003-05-01') == '0'  # Thu
    assert week_of_month('2003-12-31') == '5'
    assert week_of_month('2004-01-01') == '0'  # Thu
    assert week_of_month('2004-02-29') == '5'


def test_week_numbering_friday_mindays_1():
    locale = Locale.parse('dv_MV')
    assert locale.first_week_day == 4
    assert locale.min_week_days == 1

    def week_number(value):
        value = date.fromisoformat(value)
        return dates.format_date(value, format="YYYY-'W'ww-e", locale=locale)

    def week_of_month(value):
        value = date.fromisoformat(value)
        return dates.format_date(value, format='W', locale=locale)

    assert week_number('2003-01-01') == '2003-W01-6'
    assert week_number('2003-01-03') == '2003-W02-1'
    assert week_number('2003-12-25') == '2003-W52-7'
    assert week_number('2003-12-26') == '2004-W01-1'
    assert week_number('2003-12-31') == '2004-W01-6'
    assert week_number('2004-01-01') == '2004-W01-7'
    assert week_number('2004-01-02') == '2004-W02-1'
    assert week_number('2004-12-23') == '2004-W52-7'
    assert week_number('2004-12-24') == '2004-W53-1'
    assert week_number('2004-12-30') == '2004-W53-7'
    assert week_number('2004-12-31') == '2005-W01-1'
    assert week_number('2005-01-01') == '2005-W01-2'
    assert week_number('2005-01-07') == '2005-W02-1'
    assert week_number('2005-12-29') == '2005-W52-7'
    assert week_number('2005-12-30') == '2006-W01-1'
    assert week_number('2005-12-31') == '2006-W01-2'
    assert week_number('2006-01-01') == '2006-W01-3'
    assert week_number('2006-01-06') == '2006-W02-1'
    assert week_number('2006-12-28') == '2006-W52-7'
    assert week_number('2006-12-29') == '2007-W01-1'
    assert week_number('2006-12-31') == '2007-W01-3'
    assert week_number('2007-01-01') == '2007-W01-4'
    assert week_number('2007-01-05') == '2007-W02-1'
    assert week_number('2007-12-27') == '2007-W52-7'
    assert week_number('2007-12-28') == '2008-W01-1'
    assert week_number('2007-12-31') == '2008-W01-4'
    assert week_number('2008-01-01') == '2008-W01-5'
    assert week_number('2008-01-04') == '2008-W02-1'
    assert week_number('2008-12-25') == '2008-W52-7'
    assert week_number('2008-12-26') == '2009-W01-1'
    assert week_number('2008-12-31') == '2009-W01-6'
    assert week_number('2009-01-01') == '2009-W01-7'
    assert week_number('2009-01-02') == '2009-W02-1'
    assert week_number('2009-12-24') == '2009-W52-7'
    assert week_number('2009-12-25') == '2009-W53-1'
    assert week_number('2009-12-31') == '2009-W53-7'
    assert week_number('2010-01-01') == '2010-W01-1'
    assert week_number('2010-01-08') == '2010-W02-1'

    assert week_of_month('2003-01-01') == '1'  # Wed
    assert week_of_month('2003-02-28') == '5'
    assert week_of_month('2003-03-01') == '1'  # Sun
    assert week_of_month('2003-12-31') == '5'
    assert week_of_month('2004-01-01') == '1'  # Thu
    assert week_of_month('2004-02-29') == '5'


def test_week_numbering_saturday_mindays_1():
    locale = Locale.parse('fr_DZ')
    assert locale.first_week_day == 5
    assert locale.min_week_days == 1

    def week_number(value):
        value = date.fromisoformat(value)
        return dates.format_date(value, format="YYYY-'W'ww-e", locale=locale)

    def week_of_month(value):
        value = date.fromisoformat(value)
        return dates.format_date(value, format='W', locale=locale)

    assert week_number('2003-01-01') == '2003-W01-5'
    assert week_number('2003-01-04') == '2003-W02-1'
    assert week_number('2003-12-26') == '2003-W52-7'
    assert week_number('2003-12-27') == '2004-W01-1'
    assert week_number('2003-12-31') == '2004-W01-5'
    assert week_number('2004-01-01') == '2004-W01-6'
    assert week_number('2004-01-03') == '2004-W02-1'
    assert week_number('2004-12-24') == '2004-W52-7'
    assert week_number('2004-12-31') == '2004-W53-7'
    assert week_number('2005-01-01') == '2005-W01-1'
    assert week_number('2005-01-08') == '2005-W02-1'
    assert week_number('2005-12-30') == '2005-W52-7'
    assert week_number('2005-12-31') == '2006-W01-1'
    assert week_number('2006-01-01') == '2006-W01-2'
    assert week_number('2006-01-07') == '2006-W02-1'
    assert week_number('2006-12-29') == '2006-W52-7'
    assert week_number('2006-12-31') == '2007-W01-2'
    assert week_number('2007-01-01') == '2007-W01-3'
    assert week_number('2007-01-06') == '2007-W02-1'
    assert week_number('2007-12-28') == '2007-W52-7'
    assert week_number('2007-12-31') == '2008-W01-3'
    assert week_number('2008-01-01') == '2008-W01-4'
    assert week_number('2008-01-05') == '2008-W02-1'
    assert week_number('2008-12-26') == '2008-W52-7'
    assert week_number('2008-12-27') == '2009-W01-1'
    assert week_number('2008-12-31') == '2009-W01-5'
    assert week_number('2009-01-01') == '2009-W01-6'
    assert week_number('2009-01-03') == '2009-W02-1'
    assert week_number('2009-12-25') == '2009-W52-7'
    assert week_number('2009-12-26') == '2010-W01-1'
    assert week_number('2009-12-31') == '2010-W01-6'
    assert week_number('2010-01-01') == '2010-W01-7'
    assert week_number('2010-01-02') == '2010-W02-1'

    assert week_of_month('2003-01-01') == '1'  # Wed
    assert week_of_month('2003-02-28') == '4'
    assert week_of_month('2003-08-01') == '1'  # Fri
    assert week_of_month('2003-12-31') == '5'
    assert week_of_month('2004-01-01') == '1'  # Thu
    assert week_of_month('2004-02-29') == '5'


def test_en_gb_first_weekday():
    assert Locale.parse('en').first_week_day == 0  # Monday in general
    assert Locale.parse('en_US').first_week_day == 6  # Sunday in the US
    assert Locale.parse('en_GB').first_week_day == 0  # Monday in the UK


def test_issue_798():
    assert dates.format_timedelta(timedelta(), format='narrow', locale='es_US') == '0s'


def test_issue_892():
    assert dates.format_timedelta(timedelta(seconds=1), format='narrow', locale='pt_BR') == '1 s'
    assert dates.format_timedelta(timedelta(minutes=1), format='narrow', locale='pt_BR') == '1 min'
    assert dates.format_timedelta(timedelta(hours=1), format='narrow', locale='pt_BR') == '1 h'
    assert dates.format_timedelta(timedelta(days=1), format='narrow', locale='pt_BR') == '1 dia'
    assert dates.format_timedelta(timedelta(days=30), format='narrow', locale='pt_BR') == '1 mês'
    assert dates.format_timedelta(timedelta(days=365), format='narrow', locale='pt_BR') == '1 ano'


def test_issue_1089():
    assert dates.format_datetime(datetime.now(), locale="ja_JP@mod")
    assert dates.format_datetime(datetime.now(), locale=Locale.parse("ja_JP@mod"))


@pytest.mark.parametrize(('locale', 'format', 'negative', 'expected'), [
    ('en_US', 'long', False, 'in 3 hours'),
    ('en_US', 'long', True, '3 hours ago'),
    ('en_US', 'narrow', False, 'in 3h'),
    ('en_US', 'narrow', True, '3h ago'),
    ('en_US', 'short', False, 'in 3 hr.'),
    ('en_US', 'short', True, '3 hr. ago'),
    ('fi_FI', 'long', False, '3 tunnin päästä'),
    ('fi_FI', 'long', True, '3 tuntia sitten'),
    ('fi_FI', 'short', False, '3 t päästä'),
    ('fi_FI', 'short', True, '3 t sitten'),
    ('sv_SE', 'long', False, 'om 3 timmar'),
    ('sv_SE', 'long', True, 'för 3 timmar sedan'),
    ('sv_SE', 'short', False, 'om 3 tim'),
    ('sv_SE', 'short', True, 'för 3 tim sedan'),
])
def test_issue_1162(locale, format, negative, expected):
    delta = timedelta(seconds=10800) * (-1 if negative else +1)
    assert dates.format_timedelta(delta, add_direction=True, format=format, locale=locale) == expected



from datetime import time

import pytest

import babel.dates as dates


@pytest.mark.parametrize("locale, time, expected_period_id", [
    ("de", time(7, 42), "morning1"),  # (from, before)
    ("de", time(3, 11), "night1"),  # (after, before)
    ("fi", time(0), "midnight"),  # (at)
    ("en_US", time(12), "noon"),  # (at)
    ("en_US", time(21), "night1"),  # (from, before) across 0:00
    ("en_US", time(5), "night1"),  # (from, before) across 0:00
    ("en_US", time(6), "morning1"),  # (from, before)
    ("agq", time(10), "am"),  # no periods defined
    ("agq", time(22), "pm"),  # no periods defined
    ("am", time(14), "afternoon1"),  # (before, after)
])
def test_day_period_rules(locale, time, expected_period_id):
    assert dates.get_period_id(time, locale=locale) == expected_period_id



#
# Copyright (C) 2007-2011 Edgewall Software, 2013-2025 the Babel team
# All rights reserved.
#
# This software is licensed as described in the file LICENSE, which
# you should have received as part of this distribution. The terms
# are also available at https://github.com/python-babel/babel/blob/master/LICENSE.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at https://github.com/python-babel/babel/commits/master/.

import unittest
from datetime import datetime
from io import BytesIO, StringIO

import pytest

from babel.core import Locale
from babel.messages import pofile
from babel.messages.catalog import Catalog, Message
from babel.messages.pofile import _enclose_filename_if_necessary, _extract_locations
from babel.util import FixedOffsetTimezone


class ReadPoTestCase(unittest.TestCase):

    def test_preserve_locale(self):
        buf = StringIO(r'''msgid "foo"
msgstr "Voh"''')
        catalog = pofile.read_po(buf, locale='en_US')
        assert Locale('en', 'US') == catalog.locale

    def test_locale_gets_overridden_by_file(self):
        buf = StringIO(r'''
msgid ""
msgstr ""
"Language: en_US\n"''')
        catalog = pofile.read_po(buf, locale='de')
        assert Locale('en', 'US') == catalog.locale
        buf = StringIO(r'''
msgid ""
msgstr ""
"Language: ko-KR\n"''')
        catalog = pofile.read_po(buf, locale='de')
        assert Locale('ko', 'KR') == catalog.locale

    def test_preserve_domain(self):
        buf = StringIO(r'''msgid "foo"
msgstr "Voh"''')
        catalog = pofile.read_po(buf, domain='mydomain')
        assert catalog.domain == 'mydomain'

    def test_applies_specified_encoding_during_read(self):
        buf = BytesIO('''
msgid ""
msgstr ""
"Project-Id-Version:  3.15\\n"
"Report-Msgid-Bugs-To: Fliegender Zirkus <fliegender@zirkus.de>\\n"
"POT-Creation-Date: 2007-09-27 11:19+0700\\n"
"PO-Revision-Date: 2007-09-27 21:42-0700\\n"
"Last-Translator: John <cleese@bavaria.de>\\n"
"Language-Team: German Lang <de@babel.org>\\n"
"Plural-Forms: nplurals=2; plural=(n != 1);\\n"
"MIME-Version: 1.0\\n"
"Content-Type: text/plain; charset=iso-8859-1\\n"
"Content-Transfer-Encoding: 8bit\\n"
"Generated-By: Babel 1.0dev-r313\\n"

msgid "foo"
msgstr "bär"'''.encode('iso-8859-1'))
        catalog = pofile.read_po(buf, locale='de_DE')
        assert catalog.get('foo').string == 'bär'

    def test_encoding_header_read(self):
        buf = BytesIO(b'msgid ""\nmsgstr ""\n"Content-Type: text/plain; charset=mac_roman\\n"\n')
        catalog = pofile.read_po(buf, locale='xx_XX')
        assert catalog.charset == 'mac_roman'

    def test_plural_forms_header_parsed(self):
        buf = BytesIO(b'msgid ""\nmsgstr ""\n"Plural-Forms: nplurals=42; plural=(n % 11);\\n"\n')
        catalog = pofile.read_po(buf, locale='xx_XX')
        assert catalog.plural_expr == '(n % 11)'
        assert catalog.num_plurals == 42

    def test_read_multiline(self):
        buf = StringIO(r'''msgid ""
"Here's some text that\n"
"includesareallylongwordthatmightbutshouldnt"
" throw us into an infinite "
"loop\n"
msgstr ""''')
        catalog = pofile.read_po(buf)
        assert len(catalog) == 1
        message = list(catalog)[1]
        assert message.id == (
            "Here's some text that\nincludesareallylongwordthat"
            "mightbutshouldnt throw us into an infinite loop\n"
        )

    def test_fuzzy_header(self):
        buf = StringIO(r'''
# Translations template for AReallyReallyLongNameForAProject.
# Copyright (C) 2007 ORGANIZATION
# This file is distributed under the same license as the
# AReallyReallyLongNameForAProject project.
# FIRST AUTHOR <EMAIL@ADDRESS>, 2007.
#
#, fuzzy
''')
        catalog = pofile.read_po(buf)
        assert len(list(catalog)) == 1
        assert list(catalog)[0].fuzzy

    def test_not_fuzzy_header(self):
        buf = StringIO(r'''
# Translations template for AReallyReallyLongNameForAProject.
# Copyright (C) 2007 ORGANIZATION
# This file is distributed under the same license as the
# AReallyReallyLongNameForAProject project.
# FIRST AUTHOR <EMAIL@ADDRESS>, 2007.
#
''')
        catalog = pofile.read_po(buf)
        assert len(list(catalog)) == 1
        assert not list(catalog)[0].fuzzy

    def test_header_entry(self):
        buf = StringIO(r'''
# SOME DESCRIPTIVE TITLE.
# Copyright (C) 2007 THE PACKAGE'S COPYRIGHT HOLDER
# This file is distributed under the same license as the PACKAGE package.
# FIRST AUTHOR <EMAIL@ADDRESS>, 2007.
#
#, fuzzy
msgid ""
msgstr ""
"Project-Id-Version:  3.15\n"
"Report-Msgid-Bugs-To: Fliegender Zirkus <fliegender@zirkus.de>\n"
"POT-Creation-Date: 2007-09-27 11:19+0700\n"
"PO-Revision-Date: 2007-09-27 21:42-0700\n"
"Last-Translator: John <cleese@bavaria.de>\n"
"Language: de\n"
"Language-Team: German Lang <de@babel.org>\n"
"Plural-Forms: nplurals=2; plural=(n != 1);\n"
"MIME-Version: 1.0\n"
"Content-Type: text/plain; charset=iso-8859-2\n"
"Content-Transfer-Encoding: 8bit\n"
"Generated-By: Babel 1.0dev-r313\n"
''')
        catalog = pofile.read_po(buf)
        assert len(list(catalog)) == 1
        assert catalog.version == '3.15'
        assert catalog.msgid_bugs_address == 'Fliegender Zirkus <fliegender@zirkus.de>'
        assert datetime(2007, 9, 27, 11, 19, tzinfo=FixedOffsetTimezone(7 * 60)) == catalog.creation_date
        assert catalog.last_translator == 'John <cleese@bavaria.de>'
        assert Locale('de') == catalog.locale
        assert catalog.language_team == 'German Lang <de@babel.org>'
        assert catalog.charset == 'iso-8859-2'
        assert list(catalog)[0].fuzzy

    def test_obsolete_message(self):
        buf = StringIO(r'''# This is an obsolete message
#~ msgid "foo"
#~ msgstr "Voh"

# This message is not obsolete
#: main.py:1
msgid "bar"
msgstr "Bahr"
''')
        catalog = pofile.read_po(buf)
        assert len(catalog) == 1
        assert len(catalog.obsolete) == 1
        message = catalog.obsolete['foo']
        assert message.id == 'foo'
        assert message.string == 'Voh'
        assert message.user_comments == ['This is an obsolete message']

    def test_obsolete_message_ignored(self):
        buf = StringIO(r'''# This is an obsolete message
#~ msgid "foo"
#~ msgstr "Voh"

# This message is not obsolete
#: main.py:1
msgid "bar"
msgstr "Bahr"
''')
        catalog = pofile.read_po(buf, ignore_obsolete=True)
        assert len(catalog) == 1
        assert len(catalog.obsolete) == 0

    def test_multi_line_obsolete_message(self):
        buf = StringIO(r'''# This is an obsolete message
#~ msgid ""
#~ "foo"
#~ "foo"
#~ msgstr ""
#~ "Voh"
#~ "Vooooh"

# This message is not obsolete
#: main.py:1
msgid "bar"
msgstr "Bahr"
''')
        catalog = pofile.read_po(buf)
        assert len(catalog.obsolete) == 1
        message = catalog.obsolete['foofoo']
        assert message.id == 'foofoo'
        assert message.string == 'VohVooooh'
        assert message.user_comments == ['This is an obsolete message']

    def test_unit_following_multi_line_obsolete_message(self):
        buf = StringIO(r'''# This is an obsolete message
#~ msgid ""
#~ "foo"
#~ "fooooooo"
#~ msgstr ""
#~ "Voh"
#~ "Vooooh"

# This message is not obsolete
#: main.py:1
msgid "bar"
msgstr "Bahr"
''')
        catalog = pofile.read_po(buf)
        assert len(catalog) == 1
        message = catalog['bar']
        assert message.id == 'bar'
        assert message.string == 'Bahr'
        assert message.user_comments == ['This message is not obsolete']

    def test_unit_before_obsolete_is_not_obsoleted(self):
        buf = StringIO(r'''
# This message is not obsolete
#: main.py:1
msgid "bar"
msgstr "Bahr"

# This is an obsolete message
#~ msgid ""
#~ "foo"
#~ "fooooooo"
#~ msgstr ""
#~ "Voh"
#~ "Vooooh"
''')
        catalog = pofile.read_po(buf)
        assert len(catalog) == 1
        message = catalog['bar']
        assert message.id == 'bar'
        assert message.string == 'Bahr'
        assert message.user_comments == ['This message is not obsolete']

    def test_with_context(self):
        buf = BytesIO(b'''# Some string in the menu
#: main.py:1
msgctxt "Menu"
msgid "foo"
msgstr "Voh"

# Another string in the menu
#: main.py:2
msgctxt "Menu"
msgid "bar"
msgstr "Bahr"
''')
        catalog = pofile.read_po(buf, ignore_obsolete=True)
        assert len(catalog) == 2
        message = catalog.get('foo', context='Menu')
        assert message.context == 'Menu'
        message = catalog.get('bar', context='Menu')
        assert message.context == 'Menu'

        # And verify it pass through write_po
        out_buf = BytesIO()
        pofile.write_po(out_buf, catalog, omit_header=True)
        assert out_buf.getvalue().strip() == buf.getvalue().strip()

    def test_obsolete_message_with_context(self):
        buf = StringIO('''
# This message is not obsolete
msgid "baz"
msgstr "Bazczch"

# This is an obsolete message
#~ msgctxt "other"
#~ msgid "foo"
#~ msgstr "Voh"

# This message is not obsolete
#: main.py:1
msgid "bar"
msgstr "Bahr"
''')
        catalog = pofile.read_po(buf)
        assert len(catalog) == 2
        assert len(catalog.obsolete) == 1
        message = catalog.obsolete[("foo", "other")]
        assert message.context == 'other'
        assert message.string == 'Voh'

    def test_obsolete_messages_with_context(self):
        buf = StringIO('''
# This is an obsolete message
#~ msgctxt "apple"
#~ msgid "foo"
#~ msgstr "Foo"

# This is an obsolete message with the same id but different context
#~ msgctxt "orange"
#~ msgid "foo"
#~ msgstr "Bar"
''')
        catalog = pofile.read_po(buf)
        assert len(catalog) == 0
        assert len(catalog.obsolete) == 2
        assert 'foo' not in catalog.obsolete

        apple_msg = catalog.obsolete[('foo', 'apple')]
        assert apple_msg.id == 'foo'
        assert apple_msg.string == 'Foo'
        assert apple_msg.user_comments == ['This is an obsolete message']

        orange_msg = catalog.obsolete[('foo', 'orange')]
        assert orange_msg.id == 'foo'
        assert orange_msg.string == 'Bar'
        assert orange_msg.user_comments == ['This is an obsolete message with the same id but different context']

    def test_obsolete_messages_roundtrip(self):
        buf = StringIO('''\
# This message is not obsolete
#: main.py:1
msgid "bar"
msgstr "Bahr"

# This is an obsolete message
#~ msgid "foo"
#~ msgstr "Voh"

# This is an obsolete message
#~ msgctxt "apple"
#~ msgid "foo"
#~ msgstr "Foo"

# This is an obsolete message with the same id but different context
#~ msgctxt "orange"
#~ msgid "foo"
#~ msgstr "Bar"

''')
        generated_po_file = ''.join(pofile.generate_po(pofile.read_po(buf), omit_header=True))
        assert buf.getvalue() == generated_po_file

    def test_multiline_context(self):
        buf = StringIO('''
msgctxt "a really long "
"message context "
"why?"
msgid "mid"
msgstr "mst"
        ''')
        catalog = pofile.read_po(buf)
        assert len(catalog) == 1
        message = catalog.get('mid', context="a really long message context why?")
        assert message is not None
        assert message.context == 'a really long message context why?'

    def test_with_context_two(self):
        buf = BytesIO(b'''msgctxt "Menu"
msgid "foo"
msgstr "Voh"

msgctxt "Mannu"
msgid "bar"
msgstr "Bahr"
''')
        catalog = pofile.read_po(buf, ignore_obsolete=True)
        assert len(catalog) == 2
        message = catalog.get('foo', context='Menu')
        assert message.context == 'Menu'
        message = catalog.get('bar', context='Mannu')
        assert message.context == 'Mannu'

        # And verify it pass through write_po
        out_buf = BytesIO()
        pofile.write_po(out_buf, catalog, omit_header=True)
        assert out_buf.getvalue().strip() == buf.getvalue().strip(), out_buf.getvalue()

    def test_single_plural_form(self):
        buf = StringIO(r'''msgid "foo"
msgid_plural "foos"
msgstr[0] "Voh"''')
        catalog = pofile.read_po(buf, locale='ja_JP')
        assert len(catalog) == 1
        assert catalog.num_plurals == 1
        message = catalog['foo']
        assert len(message.string) == 1

    def test_singular_plural_form(self):
        buf = StringIO(r'''msgid "foo"
msgid_plural "foos"
msgstr[0] "Voh"
msgstr[1] "Vohs"''')
        catalog = pofile.read_po(buf, locale='nl_NL')
        assert len(catalog) == 1
        assert catalog.num_plurals == 2
        message = catalog['foo']
        assert len(message.string) == 2

    def test_more_than_two_plural_forms(self):
        buf = StringIO(r'''msgid "foo"
msgid_plural "foos"
msgstr[0] "Voh"
msgstr[1] "Vohs"
msgstr[2] "Vohss"''')
        catalog = pofile.read_po(buf, locale='lv_LV')
        assert len(catalog) == 1
        assert catalog.num_plurals == 3
        message = catalog['foo']
        assert len(message.string) == 3
        assert message.string[2] == 'Vohss'

    def test_plural_with_square_brackets(self):
        buf = StringIO(r'''msgid "foo"
msgid_plural "foos"
msgstr[0] "Voh [text]"
msgstr[1] "Vohs [text]"''')
        catalog = pofile.read_po(buf, locale='nb_NO')
        assert len(catalog) == 1
        assert catalog.num_plurals == 2
        message = catalog['foo']
        assert len(message.string) == 2

    def test_obsolete_plural_with_square_brackets(self):
        buf = StringIO('''\
#~ msgid "foo"
#~ msgid_plural "foos"
#~ msgstr[0] "Voh [text]"
#~ msgstr[1] "Vohs [text]"
''')
        catalog = pofile.read_po(buf, locale='nb_NO')
        assert len(catalog) == 0
        assert len(catalog.obsolete) == 1
        assert catalog.num_plurals == 2
        message = catalog.obsolete['foo']
        assert len(message.string) == 2
        assert message.string[0] == 'Voh [text]'
        assert message.string[1] == 'Vohs [text]'

    def test_missing_plural(self):
        buf = StringIO('''\
msgid ""
msgstr ""
"Plural-Forms: nplurals=3; plural=(n < 2) ? n : 2;\n"

msgid "foo"
msgid_plural "foos"
msgstr[0] "Voh [text]"
msgstr[1] "Vohs [text]"
''')
        catalog = pofile.read_po(buf, locale='nb_NO')
        assert len(catalog) == 1
        assert catalog.num_plurals == 3
        message = catalog['foo']
        assert len(message.string) == 3
        assert message.string[0] == 'Voh [text]'
        assert message.string[1] == 'Vohs [text]'
        assert message.string[2] == ''

    def test_missing_plural_in_the_middle(self):
        buf = StringIO('''\
msgid ""
msgstr ""
"Plural-Forms: nplurals=3; plural=(n < 2) ? n : 2;\n"

msgid "foo"
msgid_plural "foos"
msgstr[0] "Voh [text]"
msgstr[2] "Vohs [text]"
''')
        catalog = pofile.read_po(buf, locale='nb_NO')
        assert len(catalog) == 1
        assert catalog.num_plurals == 3
        message = catalog['foo']
        assert len(message.string) == 3
        assert message.string[0] == 'Voh [text]'
        assert message.string[1] == ''
        assert message.string[2] == 'Vohs [text]'

    def test_with_location(self):
        buf = StringIO('''\
#: main.py:1 \u2068filename with whitespace.py\u2069:123
msgid "foo"
msgstr "bar"
''')
        catalog = pofile.read_po(buf, locale='de_DE')
        assert len(catalog) == 1
        message = catalog['foo']
        assert message.string == 'bar'
        assert message.locations == [("main.py", 1), ("filename with whitespace.py", 123)]


    def test_abort_invalid_po_file(self):
        invalid_po = '''
            msgctxt ""
            "{\"checksum\": 2148532640, \"cxt\": \"collector_thankyou\", \"id\": "
            "270005359}"
            msgid ""
            "Thank you very much for your time.\n"
            "If you have any questions regarding this survey, please contact Fulano "
            "at nadie@blah.com"
            msgstr "Merci de prendre le temps de remplir le sondage.
            Pour toute question, veuillez communiquer avec Fulano  à nadie@blah.com
            "
        '''
        invalid_po_2 = '''
            msgctxt ""
            "{\"checksum\": 2148532640, \"cxt\": \"collector_thankyou\", \"id\": "
            "270005359}"
            msgid ""
            "Thank you very much for your time.\n"
            "If you have any questions regarding this survey, please contact Fulano "
            "at fulano@blah.com."
            msgstr "Merci de prendre le temps de remplir le sondage.
            Pour toute question, veuillez communiquer avec Fulano a fulano@blah.com
            "
            '''
        # Catalog not created, throws Unicode Error
        buf = StringIO(invalid_po)
        output = pofile.read_po(buf, locale='fr', abort_invalid=False)
        assert isinstance(output, Catalog)

        # Catalog not created, throws PoFileError
        buf = StringIO(invalid_po_2)
        with pytest.raises(pofile.PoFileError):
            pofile.read_po(buf, locale='fr', abort_invalid=True)

        # Catalog is created with warning, no abort
        buf = StringIO(invalid_po_2)
        output = pofile.read_po(buf, locale='fr', abort_invalid=False)
        assert isinstance(output, Catalog)

        # Catalog not created, aborted with PoFileError
        buf = StringIO(invalid_po_2)
        with pytest.raises(pofile.PoFileError):
            pofile.read_po(buf, locale='fr', abort_invalid=True)

    def test_invalid_pofile_with_abort_flag(self):
        parser = pofile.PoFileParser(None, abort_invalid=True)
        lineno = 10
        line = 'Algo esta mal'
        msg = 'invalid file'
        with pytest.raises(pofile.PoFileError):
            parser._invalid_pofile(line, lineno, msg)


class WritePoTestCase(unittest.TestCase):

    def test_join_locations(self):
        catalog = Catalog()
        catalog.add('foo', locations=[('main.py', 1)])
        catalog.add('foo', locations=[('utils.py', 3)])
        buf = BytesIO()
        pofile.write_po(buf, catalog, omit_header=True)
        assert buf.getvalue().strip() == b'''#: main.py:1 utils.py:3
msgid "foo"
msgstr ""'''

    def test_write_po_file_with_specified_charset(self):
        catalog = Catalog(charset='iso-8859-1')
        catalog.add('foo', 'äöü', locations=[('main.py', 1)])
        buf = BytesIO()
        pofile.write_po(buf, catalog, omit_header=False)
        po_file = buf.getvalue().strip()
        assert b'"Content-Type: text/plain; charset=iso-8859-1\\n"' in po_file
        assert 'msgstr "äöü"'.encode('iso-8859-1') in po_file

    def test_duplicate_comments(self):
        catalog = Catalog()
        catalog.add('foo', auto_comments=['A comment'])
        catalog.add('foo', auto_comments=['A comment'])
        buf = BytesIO()
        pofile.write_po(buf, catalog, omit_header=True)
        assert buf.getvalue().strip() == b'''#. A comment
msgid "foo"
msgstr ""'''

    def test_wrap_long_lines(self):
        text = """Here's some text where
white space and line breaks matter, and should

not be removed

"""
        catalog = Catalog()
        catalog.add(text, locations=[('main.py', 1)])
        buf = BytesIO()
        pofile.write_po(buf, catalog, no_location=True, omit_header=True,
                        width=42)
        assert buf.getvalue().strip() == b'''msgid ""
"Here's some text where\\n"
"white space and line breaks matter, and"
" should\\n"
"\\n"
"not be removed\\n"
"\\n"
msgstr ""'''

    def test_wrap_long_lines_with_long_word(self):
        text = """Here's some text that
includesareallylongwordthatmightbutshouldnt throw us into an infinite loop
"""
        catalog = Catalog()
        catalog.add(text, locations=[('main.py', 1)])
        buf = BytesIO()
        pofile.write_po(buf, catalog, no_location=True, omit_header=True,
                        width=32)
        assert buf.getvalue().strip() == b'''msgid ""
"Here's some text that\\n"
"includesareallylongwordthatmightbutshouldnt"
" throw us into an infinite "
"loop\\n"
msgstr ""'''

    def test_wrap_long_lines_in_header(self):
        """
        Verify that long lines in the header comment are wrapped correctly.
        """
        catalog = Catalog(project='AReallyReallyLongNameForAProject',
                          revision_date=datetime(2007, 4, 1))
        buf = BytesIO()
        pofile.write_po(buf, catalog)
        assert b'\n'.join(buf.getvalue().splitlines()[:7]) == b'''\
# Translations template for AReallyReallyLongNameForAProject.
# Copyright (C) 2007 ORGANIZATION
# This file is distributed under the same license as the
# AReallyReallyLongNameForAProject project.
# FIRST AUTHOR <EMAIL@ADDRESS>, 2007.
#
#, fuzzy'''

    def test_wrap_locations_with_hyphens(self):
        catalog = Catalog()
        catalog.add('foo', locations=[
            ('doupy/templates/base/navmenu.inc.html.py', 60),
        ])
        catalog.add('foo', locations=[
            ('doupy/templates/job-offers/helpers.html', 22),
        ])
        buf = BytesIO()
        pofile.write_po(buf, catalog, omit_header=True)
        assert buf.getvalue().strip() == b'''#: doupy/templates/base/navmenu.inc.html.py:60
#: doupy/templates/job-offers/helpers.html:22
msgid "foo"
msgstr ""'''

    def test_no_wrap_and_width_behaviour_on_comments(self):
        catalog = Catalog()
        catalog.add("Pretty dam long message id, which must really be big "
                    "to test this wrap behaviour, if not it won't work.",
                    locations=[("fake.py", n) for n in range(1, 30)])
        buf = BytesIO()
        pofile.write_po(buf, catalog, width=None, omit_header=True)
        assert buf.getvalue().lower() == b"""\
#: fake.py:1 fake.py:2 fake.py:3 fake.py:4 fake.py:5 fake.py:6 fake.py:7
#: fake.py:8 fake.py:9 fake.py:10 fake.py:11 fake.py:12 fake.py:13 fake.py:14
#: fake.py:15 fake.py:16 fake.py:17 fake.py:18 fake.py:19 fake.py:20 fake.py:21
#: fake.py:22 fake.py:23 fake.py:24 fake.py:25 fake.py:26 fake.py:27 fake.py:28
#: fake.py:29
msgid "pretty dam long message id, which must really be big to test this wrap behaviour, if not it won't work."
msgstr ""

"""
        buf = BytesIO()
        pofile.write_po(buf, catalog, width=100, omit_header=True)
        assert buf.getvalue().lower() == b"""\
#: fake.py:1 fake.py:2 fake.py:3 fake.py:4 fake.py:5 fake.py:6 fake.py:7 fake.py:8 fake.py:9 fake.py:10
#: fake.py:11 fake.py:12 fake.py:13 fake.py:14 fake.py:15 fake.py:16 fake.py:17 fake.py:18 fake.py:19
#: fake.py:20 fake.py:21 fake.py:22 fake.py:23 fake.py:24 fake.py:25 fake.py:26 fake.py:27 fake.py:28
#: fake.py:29
msgid ""
"pretty dam long message id, which must really be big to test this wrap behaviour, if not it won't"
" work."
msgstr ""

"""

    def test_pot_with_translator_comments(self):
        catalog = Catalog()
        catalog.add('foo', locations=[('main.py', 1)],
                    auto_comments=['Comment About `foo`'])
        catalog.add('bar', locations=[('utils.py', 3)],
                    user_comments=['Comment About `bar` with',
                                   'multiple lines.'])
        buf = BytesIO()
        pofile.write_po(buf, catalog, omit_header=True)
        assert buf.getvalue().strip() == b'''#. Comment About `foo`
#: main.py:1
msgid "foo"
msgstr ""

# Comment About `bar` with
# multiple lines.
#: utils.py:3
msgid "bar"
msgstr ""'''

    def test_po_with_obsolete_message(self):
        catalog = Catalog()
        catalog.add('foo', 'Voh', locations=[('main.py', 1)])
        catalog.obsolete['bar'] = Message('bar', 'Bahr',
                                          locations=[('utils.py', 3)],
                                          user_comments=['User comment'])
        buf = BytesIO()
        pofile.write_po(buf, catalog, omit_header=True)
        assert buf.getvalue().strip() == b'''#: main.py:1
msgid "foo"
msgstr "Voh"

# User comment
#~ msgid "bar"
#~ msgstr "Bahr"'''

    def test_po_with_multiline_obsolete_message(self):
        catalog = Catalog()
        catalog.add('foo', 'Voh', locations=[('main.py', 1)])
        msgid = r"""Here's a message that covers
multiple lines, and should still be handled
correctly.
"""
        msgstr = r"""Here's a message that covers
multiple lines, and should still be handled
correctly.
"""
        catalog.obsolete[msgid] = Message(msgid, msgstr,
                                          locations=[('utils.py', 3)])
        buf = BytesIO()
        pofile.write_po(buf, catalog, omit_header=True)
        assert buf.getvalue().strip() == b'''#: main.py:1
msgid "foo"
msgstr "Voh"

#~ msgid ""
#~ "Here's a message that covers\\n"
#~ "multiple lines, and should still be handled\\n"
#~ "correctly.\\n"
#~ msgstr ""
#~ "Here's a message that covers\\n"
#~ "multiple lines, and should still be handled\\n"
#~ "correctly.\\n"'''

    def test_po_with_obsolete_message_ignored(self):
        catalog = Catalog()
        catalog.add('foo', 'Voh', locations=[('main.py', 1)])
        catalog.obsolete['bar'] = Message('bar', 'Bahr',
                                          locations=[('utils.py', 3)],
                                          user_comments=['User comment'])
        buf = BytesIO()
        pofile.write_po(buf, catalog, omit_header=True, ignore_obsolete=True)
        assert buf.getvalue().strip() == b'''#: main.py:1
msgid "foo"
msgstr "Voh"'''

    def test_po_with_previous_msgid(self):
        catalog = Catalog()
        catalog.add('foo', 'Voh', locations=[('main.py', 1)],
                    previous_id='fo')
        buf = BytesIO()
        pofile.write_po(buf, catalog, omit_header=True, include_previous=True)
        assert buf.getvalue().strip() == b'''#: main.py:1
#| msgid "fo"
msgid "foo"
msgstr "Voh"'''

    def test_po_with_previous_msgid_plural(self):
        catalog = Catalog()
        catalog.add(('foo', 'foos'), ('Voh', 'Voeh'),
                    locations=[('main.py', 1)], previous_id=('fo', 'fos'))
        buf = BytesIO()
        pofile.write_po(buf, catalog, omit_header=True, include_previous=True)
        assert buf.getvalue().strip() == b'''#: main.py:1
#| msgid "fo"
#| msgid_plural "fos"
msgid "foo"
msgid_plural "foos"
msgstr[0] "Voh"
msgstr[1] "Voeh"'''

    def test_sorted_po(self):
        catalog = Catalog()
        catalog.add('bar', locations=[('utils.py', 3)],
                    user_comments=['Comment About `bar` with',
                                   'multiple lines.'])
        catalog.add(('foo', 'foos'), ('Voh', 'Voeh'),
                    locations=[('main.py', 1)])
        buf = BytesIO()
        pofile.write_po(buf, catalog, sort_output=True)
        value = buf.getvalue().strip()
        assert b'''\
# Comment About `bar` with
# multiple lines.
#: utils.py:3
msgid "bar"
msgstr ""

#: main.py:1
msgid "foo"
msgid_plural "foos"
msgstr[0] "Voh"
msgstr[1] "Voeh"''' in value
        assert value.find(b'msgid ""') < value.find(b'msgid "bar"') < value.find(b'msgid "foo"')

    def test_sorted_po_context(self):
        catalog = Catalog()
        catalog.add(('foo', 'foos'), ('Voh', 'Voeh'),
                    locations=[('main.py', 1)],
                    context='there')
        catalog.add(('foo', 'foos'), ('Voh', 'Voeh'),
                    locations=[('main.py', 1)])
        catalog.add(('foo', 'foos'), ('Voh', 'Voeh'),
                    locations=[('main.py', 1)],
                    context='here')
        buf = BytesIO()
        pofile.write_po(buf, catalog, sort_output=True)
        value = buf.getvalue().strip()
        # We expect the foo without ctx, followed by "here" foo and "there" foo
        assert b'''\
#: main.py:1
msgid "foo"
msgid_plural "foos"
msgstr[0] "Voh"
msgstr[1] "Voeh"

#: main.py:1
msgctxt "here"
msgid "foo"
msgid_plural "foos"
msgstr[0] "Voh"
msgstr[1] "Voeh"

#: main.py:1
msgctxt "there"
msgid "foo"
msgid_plural "foos"
msgstr[0] "Voh"
msgstr[1] "Voeh"''' in value

    def test_file_sorted_po(self):
        catalog = Catalog()
        catalog.add('bar', locations=[('utils.py', 3)])
        catalog.add(('foo', 'foos'), ('Voh', 'Voeh'), locations=[('main.py', 1)])
        buf = BytesIO()
        pofile.write_po(buf, catalog, sort_by_file=True)
        value = buf.getvalue().strip()
        assert value.find(b'main.py') < value.find(b'utils.py')

    def test_file_with_no_lineno(self):
        catalog = Catalog()
        catalog.add('bar', locations=[('utils.py', None)],
                    user_comments=['Comment About `bar` with',
                                   'multiple lines.'])
        buf = BytesIO()
        pofile.write_po(buf, catalog, sort_output=True)
        value = buf.getvalue().strip()
        assert b'''\
# Comment About `bar` with
# multiple lines.
#: utils.py
msgid "bar"
msgstr ""''' in value

    def test_silent_location_fallback(self):
        buf = BytesIO(b'''\
#: broken_file.py
msgid "missing line number"
msgstr ""

#: broken_file.py:broken_line_number
msgid "broken line number"
msgstr ""''')
        catalog = pofile.read_po(buf)
        assert catalog['missing line number'].locations == [('broken_file.py', None)]
        assert catalog['broken line number'].locations == []

    def test_include_lineno(self):
        catalog = Catalog()
        catalog.add('foo', locations=[('main.py', 1)])
        catalog.add('foo', locations=[('utils.py', 3)])
        buf = BytesIO()
        pofile.write_po(buf, catalog, omit_header=True, include_lineno=True)
        assert buf.getvalue().strip() == b'''#: main.py:1 utils.py:3
msgid "foo"
msgstr ""'''

    def test_no_include_lineno(self):
        catalog = Catalog()
        catalog.add('foo', locations=[('main.py', 1)])
        catalog.add('foo', locations=[('main.py', 2)])
        catalog.add('foo', locations=[('utils.py', 3)])
        buf = BytesIO()
        pofile.write_po(buf, catalog, omit_header=True, include_lineno=False)
        assert buf.getvalue().strip() == b'''#: main.py utils.py
msgid "foo"
msgstr ""'''

    def test_white_space_in_location(self):
        catalog = Catalog()
        catalog.add('foo', locations=[('main.py', 1)])
        catalog.add('foo', locations=[('utils b.py', 3)])
        buf = BytesIO()
        pofile.write_po(buf, catalog, omit_header=True, include_lineno=True)
        assert buf.getvalue().strip() == b'''#: main.py:1 \xe2\x81\xa8utils b.py\xe2\x81\xa9:3
msgid "foo"
msgstr ""'''

    def test_white_space_in_location_already_enclosed(self):
        catalog = Catalog()
        catalog.add('foo', locations=[('main.py', 1)])
        catalog.add('foo', locations=[('\u2068utils b.py\u2069', 3)])
        buf = BytesIO()
        pofile.write_po(buf, catalog, omit_header=True, include_lineno=True)
        assert buf.getvalue().strip() == b'''#: main.py:1 \xe2\x81\xa8utils b.py\xe2\x81\xa9:3
msgid "foo"
msgstr ""'''

    def test_tab_in_location(self):
        catalog = Catalog()
        catalog.add('foo', locations=[('main.py', 1)])
        catalog.add('foo', locations=[('utils\tb.py', 3)])
        buf = BytesIO()
        pofile.write_po(buf, catalog, omit_header=True, include_lineno=True)
        assert buf.getvalue().strip() == b'''#: main.py:1 \xe2\x81\xa8utils        b.py\xe2\x81\xa9:3
msgid "foo"
msgstr ""'''

    def test_tab_in_location_already_enclosed(self):
        catalog = Catalog()
        catalog.add('foo', locations=[('main.py', 1)])
        catalog.add('foo', locations=[('\u2068utils\tb.py\u2069', 3)])
        buf = BytesIO()
        pofile.write_po(buf, catalog, omit_header=True, include_lineno=True)
        assert buf.getvalue().strip() == b'''#: main.py:1 \xe2\x81\xa8utils        b.py\xe2\x81\xa9:3
msgid "foo"
msgstr ""'''


    def test_wrap_with_enclosed_file_locations(self):
        # Ensure that file names containing white space are not wrapped regardless of the --width parameter
        catalog = Catalog()
        catalog.add('foo', locations=[('\u2068test utils.py\u2069', 1)])
        catalog.add('foo', locations=[('\u2068test utils.py\u2069', 3)])
        buf = BytesIO()
        pofile.write_po(buf, catalog, omit_header=True, include_lineno=True, width=1)
        assert buf.getvalue().strip() == b'''#: \xe2\x81\xa8test utils.py\xe2\x81\xa9:1
#: \xe2\x81\xa8test utils.py\xe2\x81\xa9:3
msgid "foo"
msgstr ""'''


class RoundtripPoTestCase(unittest.TestCase):

    def test_enclosed_filenames_in_location_comment(self):
        catalog = Catalog()
        catalog.add("foo", lineno=2, locations=[("main 1.py", 1)], string="")
        catalog.add("bar", lineno=6, locations=[("other.py", 2)], string="")
        catalog.add("baz", lineno=10, locations=[("main 1.py", 3), ("other.py", 4)], string="")
        buf = BytesIO()
        pofile.write_po(buf, catalog, omit_header=True, include_lineno=True)
        buf.seek(0)
        catalog2 = pofile.read_po(buf)
        assert True is catalog.is_identical(catalog2)


class PofileFunctionsTestCase(unittest.TestCase):

    def test_unescape(self):
        escaped = '"Say:\\n  \\"hello, world!\\"\\n"'
        unescaped = 'Say:\n  "hello, world!"\n'
        assert unescaped != escaped
        assert unescaped == pofile.unescape(escaped)

    def test_unescape_of_quoted_newline(self):
        # regression test for #198
        assert pofile.unescape(r'"\\n"') == '\\n'

    def test_denormalize_on_msgstr_without_empty_first_line(self):
        # handle irregular multi-line msgstr (no "" as first line)
        # gracefully (#171)
        msgstr = '"multi-line\\n"\n" translation"'
        expected_denormalized = 'multi-line\n translation'

        assert expected_denormalized == pofile.denormalize(msgstr)
        assert expected_denormalized == pofile.denormalize(f'""\n{msgstr}')


@pytest.mark.parametrize(("line", "locations"), [
    ("\u2068file1.po\u2069", ["file1.po"]),
    ("file1.po \u2068file 2.po\u2069 file3.po", ["file1.po", "file 2.po", "file3.po"]),
    ("file1.po:1 \u2068file 2.po\u2069:2 file3.po:3", ["file1.po:1", "file 2.po:2", "file3.po:3"]),
    ("\u2068file1.po\u2069:1 \u2068file\t2.po\u2069:2 file3.po:3",
     ["file1.po:1", "file\t2.po:2", "file3.po:3"]),
    ("file1.po  file2.po", ["file1.po", "file2.po"]),
    ("file1.po \u2068\u2069 file2.po", ["file1.po", "file2.po"]),
])
def test_extract_locations_valid_location_comment(line, locations):
    assert locations == _extract_locations(line)


@pytest.mark.parametrize(("line",), [
    ("\u2068file 1.po",),
    ("file 1.po\u2069",),
    ("\u2069file 1.po\u2068",),
    ("\u2068file 1.po:1 \u2068file 2.po\u2069:2",),
    ("\u2068file 1.po\u2069:1 file 2.po\u2069:2",),
])
def test_extract_locations_invalid_location_comment(line):
    with pytest.raises(ValueError):
        _extract_locations(line)


@pytest.mark.parametrize(("filename",), [
    ("file.po",),
    ("file_a.po",),
    ("file-a.po",),
    ("file\n.po",),
    ("\u2068file.po\u2069",),
    ("\u2068file a.po\u2069",),
])
def test_enclose_filename_if_necessary_no_change(filename):
    assert filename == _enclose_filename_if_necessary(filename)


@pytest.mark.parametrize(("filename",), [
    ("file a.po",),
    ("file\ta.po",),
])
def test_enclose_filename_if_necessary_enclosed(filename):
    assert "\u2068" + filename + "\u2069" == _enclose_filename_if_necessary(filename)


def test_unknown_language_roundtrip():
    buf = StringIO(r'''
msgid ""
msgstr ""
"Language: sr_SP\n"''')
    catalog = pofile.read_po(buf)
    assert catalog.locale_identifier == 'sr_SP'
    assert not catalog.locale
    buf = BytesIO()
    pofile.write_po(buf, catalog)
    assert 'sr_SP' in buf.getvalue().decode()


def test_unknown_language_write():
    catalog = Catalog(locale='sr_SP')
    assert catalog.locale_identifier == 'sr_SP'
    assert not catalog.locale
    buf = BytesIO()
    pofile.write_po(buf, catalog)
    assert 'sr_SP' in buf.getvalue().decode()


def test_iterable_of_strings():
    """
    Test we can parse from an iterable of strings.
    """
    catalog = pofile.read_po(['msgid "foo"', b'msgstr "Voh"'], locale="en_US")
    assert catalog.locale == Locale("en", "US")
    assert catalog.get("foo").string == "Voh"


def test_issue_1087():
    buf = StringIO(r'''
msgid ""
msgstr ""
"Language: \n"
''')
    assert pofile.read_po(buf).locale is None


@pytest.mark.parametrize("case", ['msgid "foo"', 'msgid "foo"\nmsgid_plural "foos"'])
@pytest.mark.parametrize("abort_invalid", [False, True])
def test_issue_1134(case: str, abort_invalid: bool):
    buf = StringIO(case)

    if abort_invalid:
        # Catalog not created, aborted with PoFileError
        with pytest.raises(pofile.PoFileError, match="missing msgstr for msgid 'foo' on 0"):
            pofile.read_po(buf, abort_invalid=True)
    else:
        # Catalog is created with warning, no abort
        output = pofile.read_po(buf)
        assert len(output) == 1
        assert output["foo"].string in ((''), ('', ''))



#
# Copyright (C) 2007-2011 Edgewall Software, 2013-2025 the Babel team
# All rights reserved.
#
# This software is licensed as described in the file LICENSE, which
# you should have received as part of this distribution. The terms
# are also available at https://github.com/python-babel/babel/blob/master/LICENSE.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at https://github.com/python-babel/babel/commits/master/.

import codecs
import sys
import unittest
from io import BytesIO, StringIO

import pytest

from babel.messages import extract


class ExtractPythonTestCase(unittest.TestCase):

    def test_nested_calls(self):
        buf = BytesIO(b"""\
msg1 = _(i18n_arg.replace(r'\"', '"'))
msg2 = ungettext(i18n_arg.replace(r'\"', '"'), multi_arg.replace(r'\"', '"'), 2)
msg3 = ungettext("Babel", multi_arg.replace(r'\"', '"'), 2)
msg4 = ungettext(i18n_arg.replace(r'\"', '"'), "Babels", 2)
msg5 = ungettext('bunny', 'bunnies', random.randint(1, 2))
msg6 = ungettext(arg0, 'bunnies', random.randint(1, 2))
msg7 = _(hello.there)
msg8 = gettext('Rabbit')
msg9 = dgettext('wiki', model.addPage())
msg10 = dngettext(getDomain(), 'Page', 'Pages', 3)
msg11 = ngettext(
    "bunny",
    "bunnies",
    len(bunnies)
)
""")
        messages = list(extract.extract_python(buf,
                                               extract.DEFAULT_KEYWORDS.keys(),
                                               [], {}))
        assert messages == [
            (1, '_', None, []),
            (2, 'ungettext', (None, None, None), []),
            (3, 'ungettext', ('Babel', None, None), []),
            (4, 'ungettext', (None, 'Babels', None), []),
            (5, 'ungettext', ('bunny', 'bunnies', None), []),
            (6, 'ungettext', (None, 'bunnies', None), []),
            (7, '_', None, []),
            (8, 'gettext', 'Rabbit', []),
            (9, 'dgettext', ('wiki', None), []),
            (10, 'dngettext', (None, 'Page', 'Pages', None), []),
            (12, 'ngettext', ('bunny', 'bunnies', None), []),
        ]

    def test_extract_default_encoding_ascii(self):
        buf = BytesIO(b'_("a")')
        messages = list(extract.extract_python(
            buf, list(extract.DEFAULT_KEYWORDS), [], {},
        ))
        # Should work great in both py2 and py3
        assert messages == [(1, '_', 'a', [])]

    def test_extract_default_encoding_utf8(self):
        buf = BytesIO('_("☃")'.encode('UTF-8'))
        messages = list(extract.extract_python(
            buf, list(extract.DEFAULT_KEYWORDS), [], {},
        ))
        assert messages == [(1, '_', '☃', [])]

    def test_nested_comments(self):
        buf = BytesIO(b"""\
msg = ngettext('pylon',  # TRANSLATORS: shouldn't be
               'pylons', # TRANSLATORS: seeing this
               count)
""")
        messages = list(extract.extract_python(buf, ('ngettext',),
                                               ['TRANSLATORS:'], {}))
        assert messages == [(1, 'ngettext', ('pylon', 'pylons', None), [])]

    def test_comments_with_calls_that_spawn_multiple_lines(self):
        buf = BytesIO(b"""\
# NOTE: This Comment SHOULD Be Extracted
add_notice(req, ngettext("Catalog deleted.",
                         "Catalogs deleted.", len(selected)))

# NOTE: This Comment SHOULD Be Extracted
add_notice(req, _("Locale deleted."))


# NOTE: This Comment SHOULD Be Extracted
add_notice(req, ngettext("Foo deleted.", "Foos deleted.", len(selected)))

# NOTE: This Comment SHOULD Be Extracted
# NOTE: And This One Too
add_notice(req, ngettext("Bar deleted.",
                         "Bars deleted.", len(selected)))
""")
        messages = list(extract.extract_python(buf, ('ngettext', '_'), ['NOTE:'],

                                               {'strip_comment_tags': False}))
        assert messages[0] == (2, 'ngettext', ('Catalog deleted.', 'Catalogs deleted.', None), ['NOTE: This Comment SHOULD Be Extracted'])
        assert messages[1] == (6, '_', 'Locale deleted.', ['NOTE: This Comment SHOULD Be Extracted'])
        assert messages[2] == (10, 'ngettext', ('Foo deleted.', 'Foos deleted.', None), ['NOTE: This Comment SHOULD Be Extracted'])
        assert messages[3] == (14, 'ngettext', ('Bar deleted.', 'Bars deleted.', None), ['NOTE: This Comment SHOULD Be Extracted', 'NOTE: And This One Too'])

    def test_declarations(self):
        buf = BytesIO(b"""\
class gettext(object):
    pass
def render_body(context,x,y=_('Page arg 1'),z=_('Page arg 2'),**pageargs):
    pass
def ngettext(y='arg 1',z='arg 2',**pageargs):
    pass
class Meta:
    verbose_name = _('log entry')
""")
        messages = list(extract.extract_python(buf,
                                               extract.DEFAULT_KEYWORDS.keys(),
                                               [], {}))
        assert messages == [
            (3, '_', 'Page arg 1', []),
            (3, '_', 'Page arg 2', []),
            (8, '_', 'log entry', []),
        ]

    def test_multiline(self):
        buf = BytesIO(b"""\
msg1 = ngettext('pylon',
                'pylons', count)
msg2 = ngettext('elvis',
                'elvises',
                 count)
""")
        messages = list(extract.extract_python(buf, ('ngettext',), [], {}))
        assert messages == [
            (1, 'ngettext', ('pylon', 'pylons', None), []),
            (3, 'ngettext', ('elvis', 'elvises', None), []),
        ]

    def test_npgettext(self):
        buf = BytesIO(b"""\
msg1 = npgettext('Strings','pylon',
                'pylons', count)
msg2 = npgettext('Strings','elvis',
                'elvises',
                 count)
""")
        messages = list(extract.extract_python(buf, ('npgettext',), [], {}))
        assert messages == [
            (1, 'npgettext', ('Strings', 'pylon', 'pylons', None), []),
            (3, 'npgettext', ('Strings', 'elvis', 'elvises', None), []),
        ]
        buf = BytesIO(b"""\
msg = npgettext('Strings', 'pylon',  # TRANSLATORS: shouldn't be
                'pylons', # TRANSLATORS: seeing this
                count)
""")
        messages = list(extract.extract_python(buf, ('npgettext',),
                                               ['TRANSLATORS:'], {}))
        assert messages == [
            (1, 'npgettext', ('Strings', 'pylon', 'pylons', None), []),
        ]

    def test_triple_quoted_strings(self):
        buf = BytesIO(b"""\
msg1 = _('''pylons''')
msg2 = ngettext(r'''elvis''', \"\"\"elvises\"\"\", count)
msg2 = ngettext(\"\"\"elvis\"\"\", 'elvises', count)
""")
        messages = list(extract.extract_python(buf,
                                               extract.DEFAULT_KEYWORDS.keys(),
                                               [], {}))
        assert messages == [
            (1, '_', 'pylons', []),
            (2, 'ngettext', ('elvis', 'elvises', None), []),
            (3, 'ngettext', ('elvis', 'elvises', None), []),
        ]

    def test_multiline_strings(self):
        buf = BytesIO(b"""\
_('''This module provides internationalization and localization
support for your Python programs by providing an interface to the GNU
gettext message catalog library.''')
""")
        messages = list(extract.extract_python(buf,
                                               extract.DEFAULT_KEYWORDS.keys(),
                                               [], {}))
        assert messages == [
            (1, '_',
            'This module provides internationalization and localization\n'
            'support for your Python programs by providing an interface to '
            'the GNU\ngettext message catalog library.', []),
        ]

    def test_concatenated_strings(self):
        buf = BytesIO(b"""\
foobar = _('foo' 'bar')
""")
        messages = list(extract.extract_python(buf,
                                               extract.DEFAULT_KEYWORDS.keys(),
                                               [], {}))
        assert messages[0][2] == 'foobar'

    def test_unicode_string_arg(self):
        buf = BytesIO(b"msg = _(u'Foo Bar')")
        messages = list(extract.extract_python(buf, ('_',), [], {}))
        assert messages[0][2] == 'Foo Bar'

    def test_comment_tag(self):
        buf = BytesIO(b"""
# NOTE: A translation comment
msg = _(u'Foo Bar')
""")
        messages = list(extract.extract_python(buf, ('_',), ['NOTE:'], {}))
        assert messages[0][2] == 'Foo Bar'
        assert messages[0][3] == ['NOTE: A translation comment']

    def test_comment_tag_multiline(self):
        buf = BytesIO(b"""
# NOTE: A translation comment
# with a second line
msg = _(u'Foo Bar')
""")
        messages = list(extract.extract_python(buf, ('_',), ['NOTE:'], {}))
        assert messages[0][2] == 'Foo Bar'
        assert messages[0][3] == ['NOTE: A translation comment', 'with a second line']

    def test_translator_comments_with_previous_non_translator_comments(self):
        buf = BytesIO(b"""
# This shouldn't be in the output
# because it didn't start with a comment tag
# NOTE: A translation comment
# with a second line
msg = _(u'Foo Bar')
""")
        messages = list(extract.extract_python(buf, ('_',), ['NOTE:'], {}))
        assert messages[0][2] == 'Foo Bar'
        assert messages[0][3] == ['NOTE: A translation comment', 'with a second line']

    def test_comment_tags_not_on_start_of_comment(self):
        buf = BytesIO(b"""
# This shouldn't be in the output
# because it didn't start with a comment tag
# do NOTE: this will not be a translation comment
# NOTE: This one will be
msg = _(u'Foo Bar')
""")
        messages = list(extract.extract_python(buf, ('_',), ['NOTE:'], {}))
        assert messages[0][2] == 'Foo Bar'
        assert messages[0][3] == ['NOTE: This one will be']

    def test_multiple_comment_tags(self):
        buf = BytesIO(b"""
# NOTE1: A translation comment for tag1
# with a second line
msg = _(u'Foo Bar1')

# NOTE2: A translation comment for tag2
msg = _(u'Foo Bar2')
""")
        messages = list(extract.extract_python(buf, ('_',),
                                               ['NOTE1:', 'NOTE2:'], {}))
        assert messages[0][2] == 'Foo Bar1'
        assert messages[0][3] == ['NOTE1: A translation comment for tag1', 'with a second line']
        assert messages[1][2] == 'Foo Bar2'
        assert messages[1][3] == ['NOTE2: A translation comment for tag2']

    def test_two_succeeding_comments(self):
        buf = BytesIO(b"""
# NOTE: one
# NOTE: two
msg = _(u'Foo Bar')
""")
        messages = list(extract.extract_python(buf, ('_',), ['NOTE:'], {}))
        assert messages[0][2] == 'Foo Bar'
        assert messages[0][3] == ['NOTE: one', 'NOTE: two']

    def test_invalid_translator_comments(self):
        buf = BytesIO(b"""
# NOTE: this shouldn't apply to any messages
hello = 'there'

msg = _(u'Foo Bar')
""")
        messages = list(extract.extract_python(buf, ('_',), ['NOTE:'], {}))
        assert messages[0][2] == 'Foo Bar'
        assert messages[0][3] == []

    def test_invalid_translator_comments2(self):
        buf = BytesIO(b"""
# NOTE: Hi!
hithere = _('Hi there!')

# NOTE: you should not be seeing this in the .po
rows = [[v for v in range(0,10)] for row in range(0,10)]

# this (NOTE:) should not show up either
hello = _('Hello')
""")
        messages = list(extract.extract_python(buf, ('_',), ['NOTE:'], {}))
        assert messages[0][2] == 'Hi there!'
        assert messages[0][3] == ['NOTE: Hi!']
        assert messages[1][2] == 'Hello'
        assert messages[1][3] == []

    def test_invalid_translator_comments3(self):
        buf = BytesIO(b"""
# NOTE: Hi,

# there!
hithere = _('Hi there!')
""")
        messages = list(extract.extract_python(buf, ('_',), ['NOTE:'], {}))
        assert messages[0][2] == 'Hi there!'
        assert messages[0][3] == []

    def test_comment_tag_with_leading_space(self):
        buf = BytesIO(b"""
  #: A translation comment
  #: with leading spaces
msg = _(u'Foo Bar')
""")
        messages = list(extract.extract_python(buf, ('_',), [':'], {}))
        assert messages[0][2] == 'Foo Bar'
        assert messages[0][3] == [': A translation comment', ': with leading spaces']

    def test_different_signatures(self):
        buf = BytesIO(b"""
foo = _('foo', 'bar')
n = ngettext('hello', 'there', n=3)
n = ngettext(n=3, 'hello', 'there')
n = ngettext(n=3, *messages)
n = ngettext()
n = ngettext('foo')
""")
        messages = list(extract.extract_python(buf, ('_', 'ngettext'), [], {}))
        assert messages[0][2] == ('foo', 'bar')
        assert messages[1][2] == ('hello', 'there', None)
        assert messages[2][2] == (None, 'hello', 'there')
        assert messages[3][2] == (None, None)
        assert messages[4][2] is None
        assert messages[5][2] == 'foo'

    def test_utf8_message(self):
        buf = BytesIO("""
# NOTE: hello
msg = _('Bonjour à tous')
""".encode('utf-8'))
        messages = list(extract.extract_python(buf, ('_',), ['NOTE:'],
                                               {'encoding': 'utf-8'}))
        assert messages[0][2] == 'Bonjour à tous'
        assert messages[0][3] == ['NOTE: hello']

    def test_utf8_message_with_magic_comment(self):
        buf = BytesIO("""# -*- coding: utf-8 -*-
# NOTE: hello
msg = _('Bonjour à tous')
""".encode('utf-8'))
        messages = list(extract.extract_python(buf, ('_',), ['NOTE:'], {}))
        assert messages[0][2] == 'Bonjour à tous'
        assert messages[0][3] == ['NOTE: hello']

    def test_utf8_message_with_utf8_bom(self):
        buf = BytesIO(codecs.BOM_UTF8 + """
# NOTE: hello
msg = _('Bonjour à tous')
""".encode('utf-8'))
        messages = list(extract.extract_python(buf, ('_',), ['NOTE:'], {}))
        assert messages[0][2] == 'Bonjour à tous'
        assert messages[0][3] == ['NOTE: hello']

    def test_utf8_message_with_utf8_bom_and_magic_comment(self):
        buf = BytesIO(codecs.BOM_UTF8 + """# -*- coding: utf-8 -*-
# NOTE: hello
msg = _('Bonjour à tous')
""".encode('utf-8'))
        messages = list(extract.extract_python(buf, ('_',), ['NOTE:'], {}))
        assert messages[0][2] == 'Bonjour à tous'
        assert messages[0][3] == ['NOTE: hello']

    def test_utf8_bom_with_latin_magic_comment_fails(self):
        buf = BytesIO(codecs.BOM_UTF8 + """# -*- coding: latin-1 -*-
# NOTE: hello
msg = _('Bonjour à tous')
""".encode('utf-8'))
        with pytest.raises(SyntaxError):
            list(extract.extract_python(buf, ('_',), ['NOTE:'], {}))

    def test_utf8_raw_strings_match_unicode_strings(self):
        buf = BytesIO(codecs.BOM_UTF8 + """
msg = _('Bonjour à tous')
msgu = _(u'Bonjour à tous')
""".encode('utf-8'))
        messages = list(extract.extract_python(buf, ('_',), ['NOTE:'], {}))
        assert messages[0][2] == 'Bonjour à tous'
        assert messages[0][2] == messages[1][2]

    def test_extract_strip_comment_tags(self):
        buf = BytesIO(b"""\
#: This is a comment with a very simple
#: prefix specified
_('Servus')

# NOTE: This is a multiline comment with
# a prefix too
_('Babatschi')""")
        messages = list(extract.extract('python', buf, comment_tags=['NOTE:', ':'],
                                        strip_comment_tags=True))
        assert messages[0][1] == 'Servus'
        assert messages[0][2] == ['This is a comment with a very simple', 'prefix specified']
        assert messages[1][1] == 'Babatschi'
        assert messages[1][2] == ['This is a multiline comment with', 'a prefix too']

    def test_nested_messages(self):
        buf = BytesIO(b"""
# NOTE: First
_(u'Hello, {name}!', name=_(u'Foo Bar'))

# NOTE: Second
_(u'Hello, {name1} and {name2}!', name1=_(u'Heungsub'),
  name2=_(u'Armin'))

# NOTE: Third
_(u'Hello, {0} and {1}!', _(u'Heungsub'),
  _(u'Armin'))
""")
        messages = list(extract.extract_python(buf, ('_',), ['NOTE:'], {}))
        assert messages[0][2] == ('Hello, {name}!', None)
        assert messages[0][3] == ['NOTE: First']
        assert messages[1][2] == 'Foo Bar'
        assert messages[1][3] == []
        assert messages[2][2] == ('Hello, {name1} and {name2}!', None)
        assert messages[2][3] == ['NOTE: Second']
        assert messages[3][2] == 'Heungsub'
        assert messages[3][3] == []
        assert messages[4][2] == 'Armin'
        assert messages[4][3] == []
        assert messages[5][2] == ('Hello, {0} and {1}!', None)
        assert messages[5][3] == ['NOTE: Third']
        assert messages[6][2] == 'Heungsub'
        assert messages[6][3] == []
        assert messages[7][2] == 'Armin'
        assert messages[7][3] == []


class ExtractTestCase(unittest.TestCase):

    def test_invalid_filter(self):
        buf = BytesIO(b"""\
msg1 = _(i18n_arg.replace(r'\"', '"'))
msg2 = ungettext(i18n_arg.replace(r'\"', '"'), multi_arg.replace(r'\"', '"'), 2)
msg3 = ungettext("Babel", multi_arg.replace(r'\"', '"'), 2)
msg4 = ungettext(i18n_arg.replace(r'\"', '"'), "Babels", 2)
msg5 = ungettext('bunny', 'bunnies', random.randint(1, 2))
msg6 = ungettext(arg0, 'bunnies', random.randint(1, 2))
msg7 = _(hello.there)
msg8 = gettext('Rabbit')
msg9 = dgettext('wiki', model.addPage())
msg10 = dngettext(domain, 'Page', 'Pages', 3)
""")
        messages = \
            list(extract.extract('python', buf, extract.DEFAULT_KEYWORDS, [],
                                 {}))
        assert messages == [
            (5, ('bunny', 'bunnies'), [], None),
            (8, 'Rabbit', [], None),
            (10, ('Page', 'Pages'), [], None),
        ]

    def test_invalid_extract_method(self):
        buf = BytesIO(b'')
        with pytest.raises(ValueError):
            list(extract.extract('spam', buf))

    def test_different_signatures(self):
        buf = BytesIO(b"""
foo = _('foo', 'bar')
n = ngettext('hello', 'there', n=3)
n = ngettext(n=3, 'hello', 'there')
n = ngettext(n=3, *messages)
n = ngettext()
n = ngettext('foo')
""")
        messages = \
            list(extract.extract('python', buf, extract.DEFAULT_KEYWORDS, [],
                                 {}))
        assert len(messages) == 2
        assert messages[0][1] == 'foo'
        assert messages[1][1] == ('hello', 'there')

    def test_empty_string_msgid(self):
        buf = BytesIO(b"""\
msg = _('')
""")
        stderr = sys.stderr
        sys.stderr = StringIO()
        try:
            messages = \
                list(extract.extract('python', buf, extract.DEFAULT_KEYWORDS,
                                     [], {}))
            assert messages == []
            assert 'warning: Empty msgid.' in sys.stderr.getvalue()
        finally:
            sys.stderr = stderr

    def test_warn_if_empty_string_msgid_found_in_context_aware_extraction_method(self):
        buf = BytesIO(b"\nmsg = pgettext('ctxt', '')\n")
        stderr = sys.stderr
        sys.stderr = StringIO()
        try:
            messages = extract.extract('python', buf)
            assert list(messages) == []
            assert 'warning: Empty msgid.' in sys.stderr.getvalue()
        finally:
            sys.stderr = stderr

    def test_extract_allows_callable(self):
        def arbitrary_extractor(fileobj, keywords, comment_tags, options):
            return [(1, None, (), ())]
        for x in extract.extract(arbitrary_extractor, BytesIO(b"")):
            assert x[0] == 1

    def test_future(self):
        buf = BytesIO(br"""
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
nbsp = _('\xa0')
""")
        messages = list(extract.extract('python', buf,
                                        extract.DEFAULT_KEYWORDS, [], {}))
        assert messages[0][1] == '\xa0'

    def test_f_strings(self):
        buf = BytesIO(br"""
t1 = _('foobar')
t2 = _(f'spameggs' f'feast')  # should be extracted; constant parts only
t2 = _(f'spameggs' 'kerroshampurilainen')  # should be extracted (mixing f with no f)
t3 = _(f'''whoa! a '''  # should be extracted (continues on following lines)
f'flying shark'
    '... hello'
)
t4 = _(f'spameggs {t1}')  # should not be extracted
""")
        messages = list(extract.extract('python', buf, extract.DEFAULT_KEYWORDS, [], {}))
        assert len(messages) == 4
        assert messages[0][1] == 'foobar'
        assert messages[1][1] == 'spameggsfeast'
        assert messages[2][1] == 'spameggskerroshampurilainen'
        assert messages[3][1] == 'whoa! a flying shark... hello'

    def test_f_strings_non_utf8(self):
        buf = BytesIO(b"""
# -- coding: latin-1 --
t2 = _(f'\xe5\xe4\xf6' f'\xc5\xc4\xd6')
""")
        messages = list(extract.extract('python', buf, extract.DEFAULT_KEYWORDS, [], {}))
        assert len(messages) == 1
        assert messages[0][1] == 'åäöÅÄÖ'



import os
import re
from collections import Counter
from subprocess import check_output

root_path = os.path.realpath(os.path.join(os.path.dirname(__file__), '..'))

aliases = {
    re.compile("Jun Omae"): "Jun Omae",
    re.compile(r"^Hugo$"): "Hugo van Kemenade",
    re.compile(r"^Tomas R([.])?"): "Tomas R.",
}


def map_alias(name):
    for pattern, alias in aliases.items():
        if pattern.match(name):
            return alias
    return name


def get_sorted_authors_list():
    authors = check_output(['git', 'log', '--format=%aN'], cwd=root_path).decode('UTF-8')
    counts = Counter(map_alias(name) for name in authors.splitlines())
    return [author for (author, count) in counts.most_common()]


def get_authors_file_content():
    author_list = "\n".join(f"- {a}" for a in get_sorted_authors_list())

    return f'''
Babel is written and maintained by the Babel team and various contributors:

{author_list}

Babel was previously developed under the Copyright of Edgewall Software.  The
following copyright notice holds true for releases before 2013: "Copyright (c)
2007 - 2011 by Edgewall Software"

In addition to the regular contributions Babel includes a fork of Lennart
Regebro's tzlocal that originally was licensed under the CC0 license.  The
original copyright of that project is "Copyright 2013 by Lennart Regebro".
'''


def write_authors_file():
    content = get_authors_file_content()
    with open(os.path.join(root_path, 'AUTHORS'), 'w', encoding='UTF-8') as fp:
        fp.write(content)


if __name__ == '__main__':
    write_authors_file()



import pytest

from babel.units import format_unit


# New units in CLDR 46
@pytest.mark.parametrize(('unit', 'count', 'expected'), [
    ('speed-light-speed', 1, '1 světlo'),
    ('speed-light-speed', 2, '2 světla'),
    ('speed-light-speed', 5, '5 světel'),
    ('concentr-portion-per-1e9', 1, '1 částice na miliardu'),
    ('concentr-portion-per-1e9', 2, '2 částice na miliardu'),
    ('concentr-portion-per-1e9', 5, '5 částic na miliardu'),
    ('duration-night', 1, '1 noc'),
    ('duration-night', 2, '2 noci'),
    ('duration-night', 5, '5 nocí'),
])
def test_new_cldr46_units(unit, count, expected):
    assert format_unit(count, unit, locale='cs_CZ') == expected



"""
These tests do not verify any results and should not be run when
looking at improving test coverage.  They just verify that basic
operations don't fail due to odd corner cases on any locale that
we ship.
"""
import datetime
import decimal

import pytest

from babel import Locale, dates, numbers, units

NUMBERS = (
    decimal.Decimal("-33.76"),  # Negative Decimal
    decimal.Decimal("13.37"),  # Positive Decimal
    1.2 - 1.0,  # Inaccurate float
    10,  # Plain old integer
    0,  # Zero
)


@pytest.mark.all_locales
def test_smoke_dates(locale):
    locale = Locale.parse(locale)
    instant = datetime.datetime.now()
    for width in ("full", "long", "medium", "short"):
        assert dates.format_date(instant, format=width, locale=locale)
        assert dates.format_datetime(instant, format=width, locale=locale)
        assert dates.format_time(instant, format=width, locale=locale)
    # Interval test
    past = instant - datetime.timedelta(hours=23)
    assert dates.format_interval(past, instant, locale=locale)
    # Duration test - at the time of writing, all locales seem to have `short` width,
    # so let's test that.
    duration = instant - instant.replace(hour=0, minute=0, second=0)
    for granularity in ('second', 'minute', 'hour', 'day'):
        assert dates.format_timedelta(duration, granularity=granularity, format="short", locale=locale)


@pytest.mark.all_locales
def test_smoke_numbers(locale):
    locale = Locale.parse(locale)
    for number in NUMBERS:
        assert numbers.format_decimal(number, locale=locale)
        assert numbers.format_decimal(number, locale=locale, numbering_system="default")
        assert numbers.format_currency(number, "EUR", locale=locale)
        assert numbers.format_currency(number, "EUR", locale=locale, numbering_system="default")
        assert numbers.format_scientific(number, locale=locale)
        assert numbers.format_scientific(number, locale=locale, numbering_system="default")
        assert numbers.format_percent(number / 100, locale=locale)
        assert numbers.format_percent(number / 100, locale=locale, numbering_system="default")


@pytest.mark.all_locales
def test_smoke_units(locale):
    locale = Locale.parse(locale)
    for unit in ('length-meter', 'mass-kilogram', 'energy-calorie', 'volume-liter'):
        for number in NUMBERS:
            assert units.format_unit(number, measurement_unit=unit, locale=locale)
            assert units.format_unit(number, measurement_unit=unit, locale=locale, numbering_system="default")



CUSTOM_EXTRACTOR_COOKIE = "custom extractor was here"


def custom_extractor(fileobj, keywords, comment_tags, options):
    if "treat" not in options:
        raise RuntimeError(f"The custom extractor refuses to run without a delicious treat; got {options!r}")
    return [(1, next(iter(keywords)), (CUSTOM_EXTRACTOR_COOKIE,), [])]



import pathlib

import pytest

from babel.messages import frontend

jinja2 = pytest.importorskip("jinja2")

jinja2_data_path = pathlib.Path(__file__).parent / "jinja2_data"


def test_jinja2_interop(monkeypatch, tmp_path):
    """
    Test that babel can extract messages from Jinja2 templates.
    """
    monkeypatch.chdir(jinja2_data_path)
    cli = frontend.CommandLineInterface()
    pot_file = tmp_path / "messages.pot"
    cli.run(['pybabel', 'extract', '--mapping', 'mapping.cfg', '-o', str(pot_file), '.'])
    assert '"Hello, %(name)s!"' in pot_file.read_text()



import os
import shlex
import shutil
import subprocess
import sys
from pathlib import Path

import pytest

from tests.messages.consts import data_dir

Distribution = pytest.importorskip("setuptools").Distribution


@pytest.mark.parametrize("kwarg,expected", [
    ("LW_", ("LW_",)),
    ("LW_ QQ Q", ("LW_", "QQ", "Q")),
    ("yiy         aia", ("yiy", "aia")),
])
def test_extract_distutils_keyword_arg_388(kwarg, expected):
    from babel.messages import frontend, setuptools_frontend

    # This is a regression test for https://github.com/python-babel/babel/issues/388

    # Note that distutils-based commands only support a single repetition of the same argument;
    # hence `--keyword ignored` will actually never end up in the output.

    cmdline = (
        f"extract_messages --no-default-keywords --keyword ignored --keyword '{kwarg}' "
        "--input-dirs . --output-file django233.pot --add-comments Bar,Foo"
    )
    d = Distribution(attrs={
        "cmdclass": setuptools_frontend.COMMANDS,
        "script_args": shlex.split(cmdline),
    })
    d.parse_command_line()
    assert len(d.commands) == 1
    cmdinst = d.get_command_obj(d.commands[0])
    cmdinst.ensure_finalized()
    assert isinstance(cmdinst, frontend.ExtractMessages)
    assert isinstance(cmdinst, setuptools_frontend.extract_messages)
    assert set(cmdinst.keywords.keys()) == set(expected)

    # Test the comma-separated comment argument while we're at it:
    assert set(cmdinst.add_comments) == {"Bar", "Foo"}


def test_setuptools_commands(tmp_path, monkeypatch):
    """
    Smoke-tests all of the setuptools versions of the commands in turn.

    Their full functionality is tested better in `test_frontend.py`.
    """
    # Copy the test project to a temporary directory and work there
    dest = tmp_path / "dest"
    shutil.copytree(data_dir, dest)
    monkeypatch.chdir(dest)

    # When in Tox, we need to hack things a bit so as not to have the
    # sub-interpreter `sys.executable` use the tox virtualenv's Babel
    # installation, so the locale data is where we expect it to be.
    if "BABEL_TOX_INI_DIR" in os.environ:
        monkeypatch.setenv("PYTHONPATH", os.environ["BABEL_TOX_INI_DIR"])

    # Initialize an empty catalog
    subprocess.check_call([
        sys.executable,
        "setup.py",
        "init_catalog",
        "-i", os.devnull,
        "-l", "fi",
        "-d", "inited",
    ])
    po_file = Path("inited/fi/LC_MESSAGES/messages.po")
    orig_po_data = po_file.read_text()
    subprocess.check_call([
        sys.executable,
        "setup.py",
        "extract_messages",
        "-o", "extracted.pot",
    ])
    pot_file = Path("extracted.pot")
    pot_data = pot_file.read_text()
    assert "FooBar, TM" in pot_data  # should be read from setup.cfg
    assert "bugs.address@email.tld" in pot_data  # should be read from setup.cfg
    subprocess.check_call([
        sys.executable,
        "setup.py",
        "update_catalog",
        "-i", "extracted.pot",
        "-d", "inited",
    ])
    new_po_data = po_file.read_text()
    assert new_po_data != orig_po_data  # check we updated the file
    subprocess.check_call([
        sys.executable,
        "setup.py",
        "compile_catalog",
        "-d", "inited",
    ])
    assert po_file.with_suffix(".mo").exists()
